""" Utility function used by the data hierarchy configuration module """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: utility_functions.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import copy
import logging
import re
from typing import Any

from PySide6.QtCore import QEvent
from PySide6.QtGui import QMouseEvent, Qt
from PySide6.QtWidgets import QMessageBox, QStyleOptionViewItem
from cloudant import CouchDB


def is_click_within_bounds(event: QEvent,
                           option: QStyleOptionViewItem) -> bool:
  """
  Check if the click event happened within the rect area of the QStyleOptionViewItem
  Args:
    event (QEvent): Mouse event captured from the view
    option (QStyleOptionViewItem): Option send during the edit event

  Returns (bool): True/False

  """
  if event and option:
    if event.type() == QEvent.MouseButtonRelease:
      e = QMouseEvent(event)
      click_x = e.x()
      click_y = e.y()
      r = option.rect
      return (r.left() < click_x < r.left() + r.width()
              and r.top() < click_y < r.top() + r.height())
  return False


def adjust_data_hierarchy_data_to_v4(data_hierarchy_types: dict[str, Any]) -> None:
  """Correct the data hierarchy data and add missing information if the loaded data is of version < 4.0

  Args:
      data_hierarchy_types (dict[str, Any]): Data hierarchy types loaded from the database

  Returns: None
  """
  if not data_hierarchy_types:
    return None
  for type_structure in data_hierarchy_types.values():
    # Adjustments previous versions <= v3.0
    type_structure.setdefault("attachments", [])
    properties = type_structure.get("prop")
    if properties is None:
      properties = {"default": []}
      type_structure["prop"] = properties
    if not isinstance(properties, dict):
      type_structure["prop"] = {"default": properties}
    # Adjustments for v4.0
    if "prop" in type_structure:
      # replace "meta" with "prop" only if it does not exist
      type_structure.setdefault("meta", type_structure["prop"])
      del type_structure["prop"]
    if "label" in type_structure:
      type_structure.setdefault("title", type_structure["label"])
      del type_structure["label"]
    type_structure.setdefault("title", "")
  return None


def show_message(message: str,
                 icon: Any = QMessageBox.Information,
                 standard_buttons: Any = QMessageBox.Ok,
                 default_button: Any = QMessageBox.Ok) -> Any:
  """
  Displays a message to the user using QMessageBox
  Args:
    message (str): Message to be displayed
    icon (Any): Message box icon
    standard_buttons (Any): Standard buttons
    default_button (Any): Default button


  Returns: Return None if a message is empty otherwise displays the message

  """
  if message:
    msg_box = QMessageBox()
    msg_box.setWindowTitle("Data Hierarchy Editor")
    msg_box.setTextFormat(Qt.RichText)
    msg_box.setIcon(icon)
    msg_box.setText(message)
    msg_box.setStandardButtons(standard_buttons)
    msg_box.setDefaultButton(default_button)
    return msg_box.exec()
  return None


def get_next_possible_structural_level_title(existing_type_titles: Any) -> str | None:
  """
  Get the title for the next possible structural type level
  Args:
    existing_type_titles (Any): The list of titles existing in the data hierarchy document

  Returns (str|None):
    The next possible name is returned with the decimal part greater than the existing largest one
  """
  if existing_type_titles is None:
    return None
  if len(existing_type_titles) <= 0:
    return "x0"
  titles = [int(title.replace('x', '').replace('X', ''))
            for title in existing_type_titles if is_structural_level(title)]
  new_level = max(titles, default=-1)
  return f"x{new_level + 1}"


def get_db(db_name: str,
           db_user: str,
           db_pass: str,
           db_url: str,
           logger: logging.Logger) -> CouchDB | None:
  """
  Get the db instance for the test purpose
  Args:
    logger (logging): Logger instance
    db_name (str): Database instance name in CouchDB
    db_user (str): Database user-name used for CouchDB access.
    db_pass (str): Database password used for CouchDB access.
    db_url (str): Database connection URL.

  Returns (CouchDB | None):
    Connected DB instance

  """
  try:
    client = CouchDB(user=db_user,
                     auth_token=db_pass,
                     url=db_url,
                     connect=True)
  except Exception as ex:
    if logger:
      logger.error(f'Could not connect with username+password to local server, error: {ex}')
    return None
  return (client[db_name]
          if db_name in client.all_dbs() else client.create_database(db_name))


def get_types_for_display(types: list[str]) -> list[str]:
  """
  Get the types for display by converting all structural types from format: xn -> 'Structure Level n'
  Args:
    types (list[str]): List of types

  Returns: Return the list of types after converting all structural types

  """
  name_prefix = "Structure level "
  return [
    name.replace('x', name_prefix) if is_structural_level(name) else name
    for name in types
  ]


def adapt_type(title: str) -> str:
  """
  Convert only structural-type from format: 'Structure Level n' -> xn
  Args:
    title (str): Title to be adapted

  Returns: Adapted title in the needed format

  """
  return title.replace("Structure level ", "x") \
    if title and title.startswith("Structure level ") \
    else title


def is_structural_level(title: str) -> bool:
  """
  Check if the title is a structural type
  Args:
    title (str): Title to be checked

  Returns: True/False

  """
  return re.compile(r'^[Xx][0-9]+$').match(title) is not None


def generate_empty_type(displayed_title: str) -> dict[str, Any]:
  """
  Generate an empty type for creating a new data hierarchy type
  Args:
    displayed_title (str): displayed_title of the new type

  Returns: Dictionary representing a bare new type

  """
  return {
    "IRI": "",
    "title": displayed_title,
    "meta": {
      "default": generate_required_metadata()
    },
    "attachments": []
  }


def generate_required_metadata() -> list[dict[str, Any]]:
  """
  Generate a list of required metadata for creating a new data hierarchy type
  Returns (list[dict[str, Any]]): List of required metadata

  """
  return [
    {
      "name": "-name",
      "query": "What is the name of the metadata?",
      "mandatory": True
    },
    {
      "name": "-tags",
      "query": "What are the tags associated with this metadata?",
      "mandatory": False
    }
  ]


def check_data_hierarchy_types(data_hierarchy_types: dict[str, Any]) -> tuple[
  dict[str, dict[str, list[str]]],
  dict[str, list[str]],
  dict[str, dict[str, list[str]]]]:
  """
  Check the data hierarchy types to see:
    1. If all the required metadata ["-name", "-tags"] present in default-group
    2. If all the meta-data have name property
    3. If all the meta-data are unique across all groups
  Args:
    data_hierarchy_types (dict[str, Any]): Data hierarchy types loaded from the database

  Returns (tuple[
  dict[str, dict[str, list[str]]],
  dict[str, list[str]],
  dict[str, dict[str, list[str]]]
  ]):
    Empty tuple if all the mandatory metadata present under default group and all meta-data have a name and all meta-data are unique
    Otherwise returns a tuple containing:
      1. Types with metadata-groups missing required metadata
      2. Types with metadata-groups missing names
      3. Types with duplicate metadata name and list of groups where the name is duplicated

  """
  if not data_hierarchy_types:
    return {}, {}, {}
  types_with_null_name_metadata: dict[str, list[str]] = {}
  types_with_missing_metadata: dict[str, dict[str, list[str]]] = {}
  types_with_duplicate_metadata: dict[str, dict[str, list[str]]] = {}
  for type_name, type_value in data_hierarchy_types.items():
    type_name = type_name.replace("x", "Structure level ") \
      if is_structural_level(type_name) \
      else type_name
    if type_value.get("meta"):
      # Check if all required meta-data are present in the default metadata group
      set_types_missing_required_metadata(type_name, type_value, types_with_missing_metadata)
      # Check if all meta-data have a name
      set_types_without_name_in_metadata(type_name, type_value, types_with_null_name_metadata)
      # Check if all meta-data are unique across all groups
      set_types_with_duplicate_metadata(type_name, type_value, types_with_duplicate_metadata)
  return types_with_missing_metadata, types_with_null_name_metadata, types_with_duplicate_metadata


def set_types_missing_required_metadata(type_name: str,
                                        type_value: dict[str, Any],
                                        types_with_missing_metadata: dict[str, dict[str, list[str]]]) -> None:
  """
  Set types and respective groups with missing required metadata in types_with_missing_metadata.
  Args:
    type_name (str): Data hierarchy type name.
    type_value (dict[str, Any]): Data hierarchy type value.
    types_with_missing_metadata (dict[str, dict[str, list[str]]]): Type groups mapping with missing metadata.

  Returns: Nothing

  """
  if not type_name:
    return
  if default_metadata := type_value.get("meta").get("default"):  # type: ignore[union-attr]
    names_in_default_group = [item.get("name") for item in default_metadata]
    required_metadata = ["-name", "-tags"]
    for req_metadata in required_metadata:
      if req_metadata not in names_in_default_group:
        if type_name not in types_with_missing_metadata:
          types_with_missing_metadata[type_name] = {}
        if "default" not in types_with_missing_metadata[type_name]:
          types_with_missing_metadata[type_name]["default"] = []
        types_with_missing_metadata[type_name]["default"].append(req_metadata)


def set_types_without_name_in_metadata(type_name: str,
                                       type_value: dict[str, Any],
                                       types_with_null_name_metadata: dict[str, list[str]]) -> None:
  """
  Set types and respective groups containing metadata with missing name in types_with_null_name_metadata.
  Args:
    type_name (str): Data hierarchy type name.
    type_value (dict[str, Any]): Data hierarchy type value.
    types_with_null_name_metadata (dict[str, list[str]]): Type groups mapping with missing name.

  Returns: Nothing

  """
  if not type_name:
    return
  for group, metadata in type_value.get("meta").items():  # type: ignore[union-attr]
    if metadata:
      names = [item.get("name") for item in metadata]
      if not all(name and not name.isspace() for name in names):
        if type_name not in types_with_null_name_metadata:
          types_with_null_name_metadata[type_name] = []
        types_with_null_name_metadata[type_name].append(group)


def set_types_with_duplicate_metadata(type_name: str,
                                      type_value: dict[str, Any],
                                      types_with_duplicate_metadata: dict[str, dict[str, list[str]]]) -> None:
  """
  Set types and respective groups containing duplicate metadata in types_with_duplicate_metadata.
  Args:
    type_name (str): Data hierarchy type name.
    type_value (dict[str, Any]): Data hierarchy type value.
    types_with_duplicate_metadata (dict[str, dict[str, list[str]]]): Type groups mapping with duplicate metadata.

  Returns: Nothing

  """
  if not type_name:
    return
  metadata_copy = copy.deepcopy(type_value.get("meta"))
  for group, metadata in type_value.get("meta").items():  # type: ignore[union-attr]
    # Check if any duplicate metadata within the same group
    names: list[str] = list(filter(None, [item.get("name") for item in metadata] if metadata else []))
    duplicates: list[str] = [name for name in names if names.count(name) > 1]
    set_duplicates(types_with_duplicate_metadata, type_name, duplicates, group)
    metadata_copy.pop(group)  # type: ignore[union-attr]
    for neighbour_group, neighbour_metadata in metadata_copy.items():  # type: ignore[union-attr]
      # Get all duplicate names after filtering away the empty strings
      duplicates = list({item.get("name") for item in metadata}.intersection(
        [item.get("name") for item in neighbour_metadata]))
      duplicates = list(filter(None, duplicates))
      set_duplicates(types_with_duplicate_metadata, type_name, duplicates, group, neighbour_group)


def set_duplicates(types_with_duplicate_metadata: dict[str, dict[str, list[str]]],
                   type_name: str,
                   duplicates: list[str],
                   group: str,
                   neighbour_group: str = "") -> None:
  """
  Set duplicates in types_with_duplicate_metadata
  Args:
    types_with_duplicate_metadata (dict[str, dict[str, list[str]]]): Type groups with duplicate metadata.
    type_name (str): Data hierarchy type name.
    duplicates (list[str]): List of duplicate metadata names
    group (str): Group name.
    neighbour_group (str, optional): Neighbour group name.

  Returns: Nothing

  """
  for name in duplicates:
    if type_name not in types_with_duplicate_metadata:
      types_with_duplicate_metadata[type_name] = {}
    if name not in types_with_duplicate_metadata[type_name]:
      types_with_duplicate_metadata[type_name][name] = []
    if neighbour_group:
      types_with_duplicate_metadata[type_name][name].extend([group, neighbour_group])
    else:
      types_with_duplicate_metadata[type_name][name].append(group)
    # Remove duplicate groups if any exists
    types_with_duplicate_metadata[type_name][name] \
      = list(set(types_with_duplicate_metadata[type_name][name]))


def get_missing_metadata_message(types_with_missing_metadata: dict[str, dict[str, list[str]]],
                                 types_with_null_name_metadata: dict[str, list[str]],
                                 types_with_duplicate_metadata: dict[str, dict[str, list[str]]]) -> str:
  """
  Get a formatted message for missing metadata
  Args:
    types_with_duplicate_metadata (dict[str, dict[str, list[str]]]): Type groups with duplicate metadata
    types_with_null_name_metadata (dict[str, str]): Type groups with missing metadata names
    types_with_missing_metadata (dict[str, dict[str, list[str]]]): Type groups with missing metadata

  Returns (str): Html formatted message

  """
  message = ""
  if (not types_with_missing_metadata
      and not types_with_null_name_metadata
      and not types_with_duplicate_metadata):
    return message
  message += "<html>"
  message = get_missing_required_metadata_formatted_message(message, types_with_missing_metadata)
  message = get_duplicate_metadata_formatted_message(message, types_with_duplicate_metadata)
  message = get_empty_metadata_name_formatted_message(message, types_with_null_name_metadata)
  message += "</html>"
  return message


def get_empty_metadata_name_formatted_message(message: str,
                                              types_with_null_name_metadata: dict[str, list[str]]) -> str:
  """
  Get html formatted message for missing metadata.
  Args:
    message (str): Html formatted message
    types_with_null_name_metadata (dict[str, list[str]]): Type groups mapping with missing name.

  Returns: Html formatted message for missing metadata name.

  """
  if types_with_null_name_metadata:
    message += "<b>Missing metadata names:</b><ul>"
    for type_name, groups_list in types_with_null_name_metadata.items():
      for group in groups_list:
        message += (f"<li>Type: <i style=\"color:Crimson\">{type_name}</i>&nbsp;&nbsp;"
                    f"Metadata Group: <i style=\"color:Crimson\">{group}</i></li>")
    message += "</ul>"
  return message


def get_missing_required_metadata_formatted_message(message: str,
                                                    types_with_missing_metadata: dict[
                                                      str, dict[str, list[str]]]) -> str:
  """
  Get html formatted message for missing required metadata.
  Args:
    message (str): Html formatted message
    types_with_missing_metadata (dict[str, dict[str, list[str]]]): Type groups mapping with missing required metadata

  Returns: Html formatted message for missing required metadata.

  """
  if types_with_missing_metadata:
    message += "<b>Missing required metadata: </b><ul>"
    for type_name, groups in types_with_missing_metadata.items():
      for group, metadata in groups.items():
        for metadata_name in metadata:
          message += (f"<li>Type: <i style=\"color:Crimson\">{type_name}</i>&nbsp;&nbsp;"
                      f"Metadata Group: <i style=\"color:Crimson\">{group}</i>&nbsp;&nbsp;"
                      f"Metadata Name: <i style=\"color:Crimson\">{metadata_name}</i></li>")
    message += "</ul>"
  return message


def get_duplicate_metadata_formatted_message(message: str,
                                             types_with_duplicate_metadata: dict[str, dict[str, list[str]]]) -> str:
  """
  Get html formatted message for duplicate metadata.
  Args:
    message (str): Html formatted message
    types_with_duplicate_metadata (dict[str, dict[str, list[str]]]): Type groups mapping with duplicate metadata

  Returns: Html formatted message for duplicate metadata.

  """
  if types_with_duplicate_metadata:
    message += "<b>Duplicate metadata found: </b><ul>"
    for type_name, duplicate in types_with_duplicate_metadata.items():
      for metadata_name, groups in duplicate.items():
        for group in groups:
          message += (f"<li>Type: <i style=\"color:Crimson\">{type_name}</i>&nbsp;&nbsp;"
                      f"Metadata Group: <i style=\"color:Crimson\">{group}</i>&nbsp;&nbsp;"
                      f"Metadata Name: <i style=\"color:Crimson\">{metadata_name}</i></li>")
    message += "</ul>"
  return message


def can_delete_type(existing_types: list[str],
                    selected_type: str) -> bool:
  """
  Check if the selected type can be deleted
    - Non-structural types can be deleted always
    - Structural type 'x0' cannot be deleted at all, the other structural types can
    only be deleted if they are the last one in the hierarchy
  Args:
    existing_types (list[str]): List of existing types
    selected_type (str): Selected type to be deleted

  Returns (bool): True/False depending on whether the selected type can be deleted
  """
  if not selected_type:
    return False
  existing_types = [t for t in existing_types if t]
  if not is_structural_level(selected_type):
    return True
  if selected_type == 'x0':
    return False
  structural_types = list(filter(is_structural_level, existing_types))
  return (False if selected_type not in structural_types else
          max(structural_types) == selected_type)



""" Long strings and dictionaries/JSON that would obfuscate code """
from typing import Any

defaultDataHierarchy: dict[str, Any] = {
  "_id": "-dataHierarchy-",
  "-version": 4,

  "x0": {"IRI": "", "attachments": [], "title": "Projects", "icon": "", "shortcut": "space",
         "meta": {"default": [
           {"name": "-name", "query": "What is the name of the project?", "mandatory": True},
           {"name": "-tags", "query": "What are the tags associated with the project?", "mandatory": True},
           {"name": "status", "query": "What is the project status",
            "list": ["active", "paused", "passive", "finished"]},
           {"name": "objective", "query": "What is the objective?"},
           {"name": "comment", "query": "#tags comments remarks :field:value:"}
         ]}},
  "x1": {"IRI": "", "attachments": [], "title": "Folders", "icon": "", "shortcut": "",
         "meta": {"default": [
           {"name": "-name", "query": "What is the name of task?", "mandatory": True},
           {"name": "-tags", "query": "What are the tags associated with the task?", "mandatory": True},
           {"name": "comment", "query": "#tags comments remarks :field:value:"}
         ]}},
  "x2": {"IRI": "", "attachments": [], "title": "Folders", "icon": "", "shortcut": "",
         "meta": {"default": [
           {"name": "-name", "query": "What is the name of subtask?", "mandatory": True},
           {"name": "-tags", "query": "What are the tags associated with the subtask?", "mandatory": True},
           {"name": "comment", "query": "#tags comments remarks :field:value:"}
         ]}},

  "measurement": {"IRI": "", "attachments": [], "title": "Measurements", "icon": "fa5s.thermometer-half",
                  "shortcut": "m",
                  "meta": {"default": [
                    {"name": "-name", "query": "What is the name of file name?", "mandatory": True},
                    {"name": "-tags", "query": "What are the tags associated with the file name?", "mandatory": True},
                    {"name": "comment", "query": "#tags comments remarks :field:value:"},
                    {"name": "-type"},
                    {"name": "image"},
                    {"name": "#_curated"},
                    {"name": "sample", "query": "Which sample was used?", "list": "sample"},
                    {"name": "procedure", "query": "Which procedure was used?", "list": "procedure"}
                  ]}},
  "sample": {"IRI": "", "attachments": [], "title": "Samples", "icon": "fa5s.vial", "shortcut": "s",
             "meta": {"default": [
               {"name": "-name", "query": "What is the name / identifier of the sample?", "mandatory": True},
               {"name": "-tags", "query": "What are the tags associated with the sample?", "mandatory": True},
               {"name": "chemistry", "query": "What is its chemical composition?"},
               {"name": "comment", "query": "#tags comments remarks :field:value:"},
               {"name": "qrCode"}
             ]}},
  "procedure": {"IRI": "", "attachments": [], "title": "Procedures", "icon": "fa5s.list-ol", "shortcut": "p",
                "meta": {"default": [
                  {"name": "-name", "query": "What is the name / path of the procedure?", "mandatory": True},
                  {"name": "-tags", "query": "What are the tags associated with the procedure?", "mandatory": True},
                  {"name": "comment", "query": "#tags comments :field:value: e.g. #SOP_v1"},
                  {"name": "content", "query": "What is procedure (Markdown possible; autofill if file given)?"}
                ]}},
  "instrument": {"IRI": "", "attachments": [], "title": "Instruments", "icon": "ri.scales-2-line", "shortcut": "i",
                 "meta": {"default": [
                   {"name": "-name", "query": "What is the name / path of the instrument?", "mandatory": True},
                   {"name": "-tags", "query": "What are the tags associated with the instrument?", "mandatory": True},
                   {"name": "comment", "query": "#tags comments :field:value: e.g. #SOP_v1"},
                   {"name": "vendor", "query": "Who is the vendor?"}
                 ]}}
}

defaultDataHierarchyNode: dict[str, list[dict[str, str]]] = {
  "default": [
    {"name": "-name", "query": "What is the file name?"},
    {"name": "-tags"},
    {"name": "comment", "query": "#tags comments remarks :field:value:"},
    {"name": "-type"}
  ]}

defaultConfiguration: dict[str, Any] = {
  "defaultProjectGroup": "research",
  "userID": "$os.getlogin()$",
  "version": 0,
  "tableColumnsMax": 16,
  "qrPrinter": {},
  "extractorDir": "$(Path(__file__).parent/'Extractors').as_posix()$",
  "extractors": {},
  "authors": [{"first": "", "last": "", "title": "", "email": "", "orcid": "",
               "organizations": [{"organization": "", "rorid": ""}]}],
  "GUI": {},
  "projectGroups": {}
}

# level 1: type of property
#   within each: array of 3: description, default, all_choices
configurationGUI: dict[str, Any] = {
  "general": {
    "theme": ["Theme",
              "light_blue",
              ["dark_amber", "dark_blue", "dark_cyan", "dark_lightgreen", "dark_pink", "dark_purple", "dark_red", \
               "dark_teal", "dark_yellow", "light_amber", "light_blue", "light_cyan", "light_lightgreen", \
               "light_pink", "light_purple", "light_red", "light_teal", "light_yellow", "none"]],
    "loggingLevel":  ["Logging level (more->less)", "INFO", ["DEBUG", "INFO", "WARNING", "ERROR"]],
    "autosave":      ["Autosave entries in form", "Yes", ["Yes", "No"]],
    "showProjectBtn":["Show project button on top-left", "Yes", ["Yes", "No"]]
  },
  "dimensions": {
    "sidebarWidth": ["Sidebar width", 280, [220, 280, 340]],
    "maxTableColumnWidth": ["Maximum column width in tables", 400, [300, 400, 500, 600]],
    "imageSizeDetails": ["Image size in details view and form", 600, [300, 400, 500, 600]],
    "imageWidthProject": ["Image width in project view", 300, [200, 250, 300, 350, 400]],
    "maxProjectLeafHeight": ["Maximum height of item in project view", 250, [200, 250, 300, 400]],
    "widthContent": ["Width of procedures in project view", 600, [400, 500, 600, 700]],
    "docTypeOffset": ["Offset of document type in project view", 500, [400, 500, 600, 700]],
    "frameSize": ["Frame width around items in project view", 6, [4, 6, 8, 10]],
  }
}

setupTextLinux = """
### Welcome to the PASTA-ELN setup for Linux
Three components are needed for proper functioning of PASTA-ELN:
- CouchDB
- Configuration of preferences / default data hierarchy
- Example project creation

This setup will analyse and (possibly) correct these items.

If the installation is successful, manually and permanently remove the 'pastaELN.log' logfile that is in your home-directory.
"""

setupTextWindows = """
### Welcome to the PASTA-ELN setup for Windows
Four components are needed for proper functioning of PASTA-ELN:
- CouchDB
- Configuration of preferences / default data hierarchy
- Shortcut creation
- Example project creation

This setup will analyse and (possibly) correct these items.

If the installation is successful, manually and permanently remove the 'pastaELN.log' logfile that is in your home-directory (folder above "My Documents").

If an attempt fails: please follow to this [website](https://pasta-eln.github.io/pasta-eln/install.html).
"""

gitWindows = """
Do you want to install git?

Be aware, downloading the installer requires some time, depending on the internet connection.
"""

rootInstallLinux = """
Do you want to install Apache CouchDB (TM)?
If you choose yes, you will be first asked to

- choose a directory to store the data
- enter the super-user password in the new terminal that will open automatically

Be aware that downloading the installer requires time, depending on the internet connection.
"""

couchDBWindows = """
Do you want to install CouchDB?

Be aware that downloading the installer requires time, depending on the internet connection.
"""

restartPastaWindows = """
Close software now (will be done automatically in the future)

Please restart the software by
- clicking on the shortcut OR
- executing the command in a new cmd.exe window
"""

exampleDataLinux = """
Do you want to create an example project?

This step helps to verify the installation and provides an helpful example for new users.

!WARNING! This process will RESET everything and thereby DELETE EVERYTHING since you installed pastaELN.

This step usually takes up to 20sec, so please be patient. Sometimes, Linux mentions that the program
is not responding and asks if to close/wait. Please WAIT.
"""

exampleDataWindows = """
Do you want to create an example project?

This step helps to verify the installation and provides an helpful example for new users.

!WARNING! This process will RESET everything and thereby DELETE EVERYTHING since you installed pastaELN.

This step usually takes up to 1min, so please be patient.
"""

shortcuts = """
### Shortcuts:

**Ctrl+Space**: List projects

**Ctrl+M**: List measurements

**Ctrl+S**: List samples

**Ctrl+P**: List procedures

**Ctrl+I**: List instruments

**Ctrl+T**: List tags

**Ctrl+U**: List unidentified

**F2**: Test file extraction

**F5**: Synchronize

**F9**: Restart

**Ctrl+?**: Verify database integrity

**Ctrl+0**: Configuration
"""

tableHeaderHelp = """
<h4>You can add custom rows via bottom text area.</h4>

If you want to add a column:
<ul>
<li> for a normal data-field (comment, content, name, type, tags, user, date), enter this field : 'comment'
<li> to check the existence of an image: enter 'image'
<li> to check if a tag is present: "#tag", in which you replace "tag" by the tag you want to look for. "_curated" is a special tag for measurements.
<li> for information inside the metadata, use a "/": e.g. "metaVendor/fileExtension", "metaUser/stress". Capitalization is important.
</ul>
"""

allIcons = [
  "fa5s.address-book", "fa5s.address-card", "fa5s.adjust", "fa5s.align-center", "fa5s.align-justify", "fa5s.align-left",
  "fa5s.align-right", "fa5s.allergies", "fa5s.ambulance", "fa5s.american-sign-language-interpreting", "fa5s.anchor",
  "fa5s.angle-double-down", "fa5s.angle-double-left", "fa5s.angle-double-right", "fa5s.angle-double-up",
  "fa5s.angle-down", "fa5s.angle-left", "fa5s.angle-right", "fa5s.angle-up", "fa5s.archive",
  "fa5s.arrow-alt-circle-down", "fa5s.arrow-alt-circle-left", "fa5s.arrow-alt-circle-right", "fa5s.arrow-alt-circle-up",
  "fa5s.arrow-circle-down", "fa5s.arrow-circle-left", "fa5s.arrow-circle-right", "fa5s.arrow-circle-up",
  "fa5s.arrow-down", "fa5s.arrow-left", "fa5s.arrow-right", "fa5s.arrow-up", "fa5s.arrows-alt", "fa5s.arrows-alt-h",
  "fa5s.arrows-alt-v", "fa5s.assistive-listening-systems", "fa5s.asterisk", "fa5s.at", "fa5s.audio-description",
  "fa5s.backward", "fa5s.balance-scale", "fa5s.ban", "fa5s.band-aid", "fa5s.barcode", "fa5s.bars", "fa5s.baseball-ball",
  "fa5s.basketball-ball", "fa5s.bath", "fa5s.battery-empty", "fa5s.battery-full", "fa5s.battery-half",
  "fa5s.battery-quarter", "fa5s.battery-three-quarters", "fa5s.bed", "fa5s.beer", "fa5s.bell", "fa5s.bell-slash",
  "fa5s.bicycle", "fa5s.binoculars", "fa5s.birthday-cake", "fa5s.blind", "fa5s.bold", "fa5s.bolt", "fa5s.bomb",
  "fa5s.book", "fa5s.bookmark", "fa5s.bowling-ball", "fa5s.box", "fa5s.box-open", "fa5s.boxes", "fa5s.braille",
  "fa5s.briefcase", "fa5s.briefcase-medical", "fa5s.bug", "fa5s.building", "fa5s.bullhorn", "fa5s.bullseye",
  "fa5s.burn", "fa5s.bus", "fa5s.calculator", "fa5s.calendar", "fa5s.calendar-alt", "fa5s.calendar-check",
  "fa5s.calendar-minus", "fa5s.calendar-plus", "fa5s.calendar-times", "fa5s.camera", "fa5s.camera-retro",
  "fa5s.capsules", "fa5s.car", "fa5s.caret-down", "fa5s.caret-left", "fa5s.caret-right", "fa5s.caret-square-down",
  "fa5s.caret-square-left", "fa5s.caret-square-right", "fa5s.caret-square-up", "fa5s.caret-up", "fa5s.cart-arrow-down",
  "fa5s.cart-plus", "fa5s.certificate", "fa5s.chart-area", "fa5s.chart-bar", "fa5s.chart-line", "fa5s.chart-pie",
  "fa5s.check", "fa5s.check-circle", "fa5s.check-square", "fa5s.chess", "fa5s.chess-bishop", "fa5s.chess-board",
  "fa5s.chess-king", "fa5s.chess-knight", "fa5s.chess-pawn", "fa5s.chess-queen", "fa5s.chess-rook",
  "fa5s.chevron-circle-down", "fa5s.chevron-circle-left", "fa5s.chevron-circle-right", "fa5s.chevron-circle-up",
  "fa5s.chevron-down", "fa5s.chevron-left", "fa5s.chevron-right", "fa5s.chevron-up", "fa5s.child", "fa5s.circle",
  "fa5s.circle-notch", "fa5s.clipboard", "fa5s.clipboard-check", "fa5s.clipboard-list", "fa5s.clock", "fa5s.clone",
  "fa5s.closed-captioning", "fa5s.cloud", "fa5s.cloud-download-alt", "fa5s.cloud-upload-alt", "fa5s.code",
  "fa5s.code-branch", "fa5s.coffee", "fa5s.cog", "fa5s.cogs", "fa5s.columns", "fa5s.comment", "fa5s.comment-alt",
  "fa5s.comment-dots", "fa5s.comment-slash", "fa5s.comments", "fa5s.compass", "fa5s.compress", "fa5s.copy",
  "fa5s.copyright", "fa5s.couch", "fa5s.credit-card", "fa5s.crop", "fa5s.crosshairs", "fa5s.cube", "fa5s.cubes",
  "fa5s.cut", "fa5s.database", "fa5s.deaf", "fa5s.desktop", "fa5s.diagnoses", "fa5s.dna", "fa5s.dollar-sign",
  "fa5s.dolly", "fa5s.dolly-flatbed", "fa5s.donate", "fa5s.dot-circle", "fa5s.dove", "fa5s.download", "fa5s.edit",
  "fa5s.eject", "fa5s.ellipsis-h", "fa5s.ellipsis-v", "fa5s.envelope", "fa5s.envelope-open", "fa5s.envelope-square",
  "fa5s.eraser", "fa5s.euro-sign", "fa5s.exchange-alt", "fa5s.exclamation", "fa5s.exclamation-circle",
  "fa5s.exclamation-triangle", "fa5s.expand", "fa5s.expand-arrows-alt", "fa5s.external-link-alt",
  "fa5s.external-link-square-alt", "fa5s.eye", "fa5s.eye-dropper", "fa5s.eye-slash", "fa5s.fast-backward",
  "fa5s.fast-forward", "fa5s.fax", "fa5s.female", "fa5s.fighter-jet", "fa5s.file", "fa5s.file-alt", "fa5s.file-archive",
  "fa5s.file-audio", "fa5s.file-code", "fa5s.file-excel", "fa5s.file-image", "fa5s.file-medical",
  "fa5s.file-medical-alt", "fa5s.file-pdf", "fa5s.file-powerpoint", "fa5s.file-video", "fa5s.file-word", "fa5s.film",
  "fa5s.filter", "fa5s.fire", "fa5s.fire-extinguisher", "fa5s.first-aid", "fa5s.flag", "fa5s.flag-checkered",
  "fa5s.flask", "fa5s.folder", "fa5s.folder-open", "fa5s.font", "fa5s.football-ball", "fa5s.forward", "fa5s.frown",
  "fa5s.futbol", "fa5s.gamepad", "fa5s.gavel", "fa5s.gem", "fa5s.genderless", "fa5s.gift", "fa5s.glass-martini",
  "fa5s.globe", "fa5s.golf-ball", "fa5s.graduation-cap", "fa5s.h-square", "fa5s.hand-holding",
  "fa5s.hand-holding-heart", "fa5s.hand-holding-usd", "fa5s.hand-lizard", "fa5s.hand-paper", "fa5s.hand-peace",
  "fa5s.hand-point-down", "fa5s.hand-point-left", "fa5s.hand-point-right", "fa5s.hand-point-up", "fa5s.hand-pointer",
  "fa5s.hand-rock", "fa5s.hand-scissors", "fa5s.hand-spock", "fa5s.hands", "fa5s.hands-helping", "fa5s.handshake",
  "fa5s.hashtag", "fa5s.hdd", "fa5s.heading", "fa5s.headphones", "fa5s.heart", "fa5s.heartbeat", "fa5s.history",
  "fa5s.hockey-puck", "fa5s.home", "fa5s.hospital", "fa5s.hospital-alt", "fa5s.hospital-symbol", "fa5s.hourglass",
  "fa5s.hourglass-end", "fa5s.hourglass-half", "fa5s.hourglass-start", "fa5s.i-cursor", "fa5s.id-badge", "fa5s.id-card",
  "fa5s.id-card-alt", "fa5s.image", "fa5s.images", "fa5s.inbox", "fa5s.indent", "fa5s.industry", "fa5s.info",
  "fa5s.info-circle", "fa5s.italic", "fa5s.key", "fa5s.keyboard", "fa5s.language", "fa5s.laptop", "fa5s.leaf",
  "fa5s.lemon", "fa5s.level-down-alt", "fa5s.level-up-alt", "fa5s.life-ring", "fa5s.lightbulb", "fa5s.link",
  "fa5s.lira-sign", "fa5s.list", "fa5s.list-alt", "fa5s.list-ol", "fa5s.list-ul", "fa5s.location-arrow", "fa5s.lock",
  "fa5s.lock-open", "fa5s.long-arrow-alt-down", "fa5s.long-arrow-alt-left", "fa5s.long-arrow-alt-right",
  "fa5s.long-arrow-alt-up", "fa5s.low-vision", "fa5s.magic", "fa5s.magnet", "fa5s.male", "fa5s.map", "fa5s.map-marker",
  "fa5s.map-marker-alt", "fa5s.map-pin", "fa5s.map-signs", "fa5s.mars", "fa5s.mars-double", "fa5s.mars-stroke",
  "fa5s.mars-stroke-h", "fa5s.mars-stroke-v", "fa5s.medkit", "fa5s.meh", "fa5s.mercury", "fa5s.microchip",
  "fa5s.microphone", "fa5s.microphone-slash", "fa5s.minus", "fa5s.minus-circle", "fa5s.minus-square", "fa5s.mobile",
  "fa5s.mobile-alt", "fa5s.money-bill-alt", "fa5s.moon", "fa5s.motorcycle", "fa5s.mouse-pointer", "fa5s.music",
  "fa5s.neuter", "fa5s.newspaper", "fa5s.notes-medical", "fa5s.object-group", "fa5s.object-ungroup", "fa5s.outdent",
  "fa5s.paint-brush", "fa5s.pallet", "fa5s.paper-plane", "fa5s.paperclip", "fa5s.parachute-box", "fa5s.paragraph",
  "fa5s.paste", "fa5s.pause", "fa5s.pause-circle", "fa5s.paw", "fa5s.pen-square", "fa5s.pencil-alt",
  "fa5s.people-carry", "fa5s.percent", "fa5s.phone", "fa5s.phone-slash", "fa5s.phone-square", "fa5s.phone-volume",
  "fa5s.piggy-bank", "fa5s.pills", "fa5s.plane", "fa5s.play", "fa5s.play-circle", "fa5s.plug", "fa5s.plus",
  "fa5s.plus-circle", "fa5s.plus-square", "fa5s.podcast", "fa5s.poo", "fa5s.pound-sign", "fa5s.power-off",
  "fa5s.prescription-bottle", "fa5s.prescription-bottle-alt", "fa5s.print", "fa5s.procedures", "fa5s.puzzle-piece",
  "fa5s.qrcode", "fa5s.question", "fa5s.question-circle", "fa5s.quidditch", "fa5s.quote-left", "fa5s.quote-right",
  "fa5s.random", "fa5s.recycle", "fa5s.redo", "fa5s.redo-alt", "fa5s.registered", "fa5s.reply", "fa5s.reply-all",
  "fa5s.retweet", "fa5s.ribbon", "fa5s.road", "fa5s.rocket", "fa5s.rss", "fa5s.rss-square", "fa5s.ruble-sign",
  "fa5s.rupee-sign", "fa5s.save", "fa5s.search", "fa5s.search-minus", "fa5s.search-plus", "fa5s.seedling",
  "fa5s.server", "fa5s.share", "fa5s.share-alt", "fa5s.share-alt-square", "fa5s.share-square", "fa5s.shekel-sign",
  "fa5s.shield-alt", "fa5s.ship", "fa5s.shipping-fast", "fa5s.shopping-bag", "fa5s.shopping-basket",
  "fa5s.shopping-cart", "fa5s.shower", "fa5s.sign", "fa5s.sign-in-alt", "fa5s.sign-language", "fa5s.sign-out-alt",
  "fa5s.signal", "fa5s.sitemap", "fa5s.sliders-h", "fa5s.smile", "fa5s.smoking", "fa5s.snowflake", "fa5s.sort",
  "fa5s.sort-alpha-down", "fa5s.sort-alpha-up", "fa5s.sort-amount-down", "fa5s.sort-amount-up", "fa5s.sort-down",
  "fa5s.sort-numeric-down", "fa5s.sort-numeric-up", "fa5s.sort-up", "fa5s.space-shuttle", "fa5s.spinner", "fa5s.square",
  "fa5s.square-full", "fa5s.star", "fa5s.star-half", "fa5s.step-backward", "fa5s.step-forward", "fa5s.stethoscope",
  "fa5s.sticky-note", "fa5s.stop", "fa5s.stop-circle", "fa5s.stopwatch", "fa5s.street-view", "fa5s.strikethrough",
  "fa5s.subscript", "fa5s.subway", "fa5s.suitcase", "fa5s.sun", "fa5s.superscript", "fa5s.sync", "fa5s.sync-alt",
  "fa5s.syringe", "fa5s.table", "fa5s.table-tennis", "fa5s.tablet", "fa5s.tablet-alt", "fa5s.tablets",
  "fa5s.tachometer-alt", "fa5s.tag", "fa5s.tags", "fa5s.tape", "fa5s.tasks", "fa5s.taxi", "fa5s.terminal",
  "fa5s.text-height", "fa5s.text-width", "fa5s.th", "fa5s.th-large", "fa5s.th-list", "fa5s.thermometer",
  "fa5s.thermometer-empty", "fa5s.thermometer-full", "fa5s.thermometer-half", "fa5s.thermometer-quarter",
  "fa5s.thermometer-three-quarters", "fa5s.thumbs-down", "fa5s.thumbs-up", "fa5s.thumbtack", "fa5s.ticket-alt",
  "fa5s.times", "fa5s.times-circle", "fa5s.tint", "fa5s.toggle-off", "fa5s.toggle-on", "fa5s.trademark", "fa5s.train",
  "fa5s.transgender", "fa5s.transgender-alt", "fa5s.trash", "fa5s.trash-alt", "fa5s.tree", "fa5s.trophy", "fa5s.truck",
  "fa5s.truck-loading", "fa5s.truck-moving", "fa5s.tty", "fa5s.tv", "fa5s.umbrella", "fa5s.underline", "fa5s.undo",
  "fa5s.undo-alt", "fa5s.universal-access", "fa5s.university", "fa5s.unlink", "fa5s.unlock", "fa5s.unlock-alt",
  "fa5s.upload", "fa5s.user", "fa5s.user-circle", "fa5s.user-md", "fa5s.user-plus", "fa5s.user-secret",
  "fa5s.user-times", "fa5s.users", "fa5s.utensil-spoon", "fa5s.utensils", "fa5s.venus", "fa5s.venus-double",
  "fa5s.venus-mars", "fa5s.vial", "fa5s.vials", "fa5s.video", "fa5s.video-slash", "fa5s.volleyball-ball",
  "fa5s.volume-down", "fa5s.volume-off", "fa5s.volume-up", "fa5s.warehouse", "fa5s.weight", "fa5s.wheelchair",
  "fa5s.wifi", "fa5s.window-close", "fa5s.window-maximize", "fa5s.window-minimize", "fa5s.window-restore",
  "fa5s.wine-glass", "fa5s.won-sign", "fa5s.wrench", "fa5s.x-ray", "fa5s.yen-sign", "far fa-address-book",
  "far fa-address-card", "far fa-arrow-alt-circle-down", "far fa-arrow-alt-circle-left",
  "far fa-arrow-alt-circle-right", "far fa-arrow-alt-circle-up", "far fa-bell", "far fa-bell-slash", "far fa-bookmark",
  "far fa-building", "far fa-calendar", "far fa-calendar-alt", "far fa-calendar-check", "far fa-calendar-minus",
  "far fa-calendar-plus", "far fa-calendar-times", "far fa-caret-square-down", "far fa-caret-square-left",
  "far fa-caret-square-right", "far fa-caret-square-up", "far fa-chart-bar", "far fa-check-circle",
  "far fa-check-square", "far fa-circle", "far fa-clipboard", "far fa-clock", "far fa-clone",
  "far fa-closed-captioning", "far fa-comment", "far fa-comment-alt", "far fa-comments", "far fa-compass",
  "far fa-copy", "far fa-copyright", "far fa-credit-card", "far fa-dot-circle", "far fa-edit", "far fa-envelope",
  "far fa-envelope-open", "far fa-eye-slash", "far fa-file", "far fa-file-alt", "far fa-file-archive",
  "far fa-file-audio", "far fa-file-code", "far fa-file-excel", "far fa-file-image", "far fa-file-pdf",
  "far fa-file-powerpoint", "far fa-file-video", "far fa-file-word", "far fa-flag", "far fa-folder",
  "far fa-folder-open", "far fa-frown", "far fa-futbol", "far fa-gem", "far fa-hand-lizard", "far fa-hand-paper",
  "far fa-hand-peace", "far fa-hand-point-down", "far fa-hand-point-left", "far fa-hand-point-right",
  "far fa-hand-point-up", "far fa-hand-pointer", "far fa-hand-rock", "far fa-hand-scissors", "far fa-hand-spock",
  "far fa-handshake", "far fa-hdd", "far fa-heart", "far fa-hospital", "far fa-hourglass", "far fa-id-badge",
  "far fa-id-card", "far fa-image", "far fa-images", "far fa-keyboard", "far fa-lemon", "far fa-life-ring",
  "far fa-lightbulb", "far fa-list-alt", "far fa-map", "far fa-meh", "far fa-minus-square", "far fa-money-bill-alt",
  "far fa-moon", "far fa-newspaper", "far fa-object-group", "far fa-object-ungroup", "far fa-paper-plane",
  "far fa-pause-circle", "far fa-play-circle", "far fa-plus-square", "far fa-question-circle", "far fa-registered",
  "far fa-save", "far fa-share-square", "far fa-smile", "far fa-snowflake", "far fa-square", "far fa-star",
  "far fa-star-half", "far fa-sticky-note", "far fa-stop-circle", "far fa-sun", "far fa-thumbs-down",
  "far fa-thumbs-up", "far fa-times-circle", "far fa-trash-alt", "far fa-user", "far fa-user-circle",
  "far fa-window-close", "far fa-window-maximize", "far fa-window-minimize", "far fa-window-restore", "fab fa-500px",
  "fab fa-accessible-icon", "fab fa-accusoft", "fab fa-adn", "fab fa-adversal", "fab fa-affiliatetheme",
  "fab fa-algolia", "fab fa-amazon", "fab fa-amazon-pay", "fab fa-amilia", "fab fa-android", "fab fa-angellist",
  "fab fa-angrycreative", "fab fa-angular", "fab fa-app-store", "fab fa-app-store-ios", "fab fa-apper", "fab fa-apple",
  "fab fa-apple-pay", "fab fa-asymmetrik", "fab fa-audible", "fab fa-autoprefixer", "fab fa-avianex", "fab fa-aviato",
  "fab fa-aws", "fab fa-bandcamp", "fab fa-behance", "fab fa-behance-square", "fab fa-bimobject", "fab fa-bitbucket",
  "fab fa-bitcoin", "fab fa-bity", "fab fa-black-tie", "fab fa-blackberry", "fab fa-blogger", "fab fa-blogger-b",
  "fab fa-bluetooth", "fab fa-bluetooth-b", "fab fa-btc", "fab fa-buromobelexperte", "fab fa-buysellads",
  "fab fa-cc-amazon-pay", "fab fa-cc-amex", "fab fa-cc-apple-pay", "fab fa-cc-diners-club", "fab fa-cc-discover",
  "fab fa-cc-jcb", "fab fa-cc-mastercard", "fab fa-cc-paypal", "fab fa-cc-stripe", "fab fa-cc-visa",
  "fab fa-centercode", "fab fa-chrome", "fab fa-cloudscale", "fab fa-cloudsmith", "fab fa-cloudversify",
  "fab fa-codepen", "fab fa-codiepie", "fab fa-connectdevelop", "fab fa-contao", "fab fa-cpanel",
  "fab fa-creative-commons", "fab fa-css3", "fab fa-css3-alt", "fab fa-cuttlefish", "fab fa-d-and-d", "fab fa-dashcube",
  "fab fa-delicious", "fab fa-deploydog", "fab fa-deskpro", "fab fa-deviantart", "fab fa-digg", "fab fa-digital-ocean",
  "fab fa-discord", "fab fa-discourse", "fab fa-dochub", "fab fa-docker", "fab fa-draft2digital", "fab fa-dribbble",
  "fab fa-dribbble-square", "fab fa-dropbox", "fab fa-drupal", "fab fa-dyalog", "fab fa-earlybirds", "fab fa-edge",
  "fab fa-elementor", "fab fa-ember", "fab fa-empire", "fab fa-envira", "fab fa-erlang", "fab fa-ethereum",
  "fab fa-etsy", "fab fa-expeditedssl", "fab fa-facebook", "fab fa-facebook-f", "fab fa-facebook-messenger",
  "fab fa-facebook-square", "fab fa-firefox", "fab fa-first-order", "fab fa-firstdraft", "fab fa-flickr",
  "fab fa-flipboard", "fab fa-fly", "fab fa-font-awesome", "fab fa-font-awesome-alt", "fab fa-font-awesome-flag",
  "fab fa-fonticons", "fab fa-fonticons-fi", "fab fa-fort-awesome", "fab fa-fort-awesome-alt", "fab fa-forumbee",
  "fab fa-foursquare", "fab fa-free-code-camp", "fab fa-freebsd", "fab fa-get-pocket", "fab fa-gg", "fab fa-gg-circle",
  "fab fa-git", "fab fa-git-square", "fab fa-github", "fab fa-github-alt", "fab fa-github-square", "fab fa-gitkraken",
  "fab fa-gitlab", "fab fa-gitter", "fab fa-glide", "fab fa-glide-g", "fab fa-gofore", "fab fa-goodreads",
  "fab fa-goodreads-g", "fab fa-google", "fab fa-google-drive", "fab fa-google-play", "fab fa-google-plus",
  "fab fa-google-plus-g", "fab fa-google-plus-square", "fab fa-google-wallet", "fab fa-gratipay", "fab fa-grav",
  "fab fa-gripfire", "fab fa-grunt", "fab fa-gulp", "fab fa-hacker-news", "fab fa-hacker-news-square", "fab fa-hips",
  "fab fa-hire-a-helper", "fab fa-hooli", "fab fa-hotjar", "fab fa-houzz", "fab fa-html5", "fab fa-hubspot",
  "fab fa-imdb", "fab fa-instagram", "fab fa-internet-explorer", "fab fa-ioxhost", "fab fa-itunes",
  "fab fa-itunes-note", "fab fa-jenkins", "fab fa-joget", "fab fa-joomla", "fab fa-js", "fab fa-js-square",
  "fab fa-jsfiddle", "fab fa-keycdn", "fab fa-kickstarter", "fab fa-kickstarter-k", "fab fa-korvue", "fab fa-laravel",
  "fab fa-lastfm", "fab fa-lastfm-square", "fab fa-leanpub", "fab fa-less", "fab fa-line", "fab fa-linkedin",
  "fab fa-linkedin-in", "fab fa-linode", "fab fa-linux", "fab fa-lyft", "fab fa-magento", "fab fa-maxcdn",
  "fab fa-medapps", "fab fa-medium", "fab fa-medium-m", "fab fa-medrt", "fab fa-meetup", "fab fa-microsoft",
  "fab fa-mix", "fab fa-mixcloud", "fab fa-mizuni", "fab fa-modx", "fab fa-monero", "fab fa-napster",
  "fab fa-nintendo-switch", "fab fa-node", "fab fa-node-js", "fab fa-npm", "fab fa-ns8", "fab fa-nutritionix",
  "fab fa-odnoklassniki", "fab fa-odnoklassniki-square", "fab fa-opencart", "fab fa-openid", "fab fa-opera",
  "fab fa-optin-monster", "fab fa-osi", "fab fa-page4", "fab fa-pagelines", "fab fa-palfed", "fab fa-patreon",
  "fab fa-paypal", "fab fa-periscope", "fab fa-phabricator", "fab fa-phoenix-framework", "fab fa-php",
  "fab fa-pied-piper", "fab fa-pied-piper-alt", "fab fa-pied-piper-pp", "fab fa-pinterest", "fab fa-pinterest-p",
  "fab fa-pinterest-square", "fab fa-playstation", "fab fa-product-hunt", "fab fa-pushed", "fab fa-python", "fab fa-qq",
  "fab fa-quinscape", "fab fa-quora", "fab fa-ravelry", "fab fa-react", "fab fa-readme", "fab fa-rebel",
  "fab fa-red-river", "fab fa-reddit", "fab fa-reddit-alien", "fab fa-reddit-square", "fab fa-rendact", "fab fa-renren",
  "fab fa-replyd", "fab fa-resolving", "fab fa-rocketchat", "fab fa-rockrms", "fab fa-safari", "fab fa-sass",
  "fab fa-schlix", "fab fa-scribd", "fab fa-searchengin", "fab fa-sellcast", "fab fa-sellsy", "fab fa-servicestack",
  "fab fa-shirtsinbulk", "fab fa-simplybuilt", "fab fa-sistrix", "fab fa-skyatlas", "fab fa-skype", "fab fa-slack",
  "fab fa-slack-hash", "fab fa-slideshare", "fab fa-snapchat", "fab fa-snapchat-ghost", "fab fa-snapchat-square",
  "fab fa-soundcloud", "fab fa-speakap", "fab fa-spotify", "fab fa-stack-exchange", "fab fa-stack-overflow",
  "fab fa-staylinked", "fab fa-steam", "fab fa-steam-square", "fab fa-steam-symbol", "fab fa-sticker-mule",
  "fab fa-strava", "fab fa-stripe", "fab fa-stripe-s", "fab fa-studiovinari", "fab fa-stumbleupon",
  "fab fa-stumbleupon-circle", "fab fa-superpowers", "fab fa-supple", "fab fa-telegram", "fab fa-telegram-plane",
  "fab fa-tencent-weibo", "fab fa-themeisle", "fab fa-trello", "fab fa-tripadvisor", "fab fa-tumblr",
  "fab fa-tumblr-square", "fab fa-twitch", "fab fa-twitter", "fab fa-twitter-square", "fab fa-typo3", "fab fa-uber",
  "fab fa-uikit", "fab fa-uniregistry", "fab fa-untappd", "fab fa-usb", "fab fa-ussunnah", "fab fa-vaadin",
  "fab fa-viacoin", "fab fa-viadeo", "fab fa-viadeo-square", "fab fa-viber", "fab fa-vimeo", "fab fa-vimeo-square",
  "fab fa-vimeo-v", "fab fa-vine", "fab fa-vk", "fab fa-vnv", "fab fa-vuejs", "fab fa-weibo", "fab fa-weixin",
  "fab fa-whatsapp", "fab fa-whatsapp-square", "fab fa-whmcs", "fab fa-wikipedia-w", "fab fa-windows",
  "fab fa-wordpress", "fab fa-wordpress-simple", "fab fa-wpbeginner", "fab fa-wpexplorer", "fab fa-wpforms",
  "fab fa-xbox", "fab fa-xing", "fab fa-xing-square", "fab fa-y-combinator", "fab fa-yahoo", "fab fa-yandex",
  "fab fa-yandex-international", "fab fa-yelp", "fab fa-yoast", "fab fa-youtube", "fab fa-youtube-square"
]



""" Python Backend: all operations with the filesystem are here """
import json, sys, os, importlib, tempfile, logging, traceback
from threading import Thread
from pathlib import Path
from typing import Any, Optional, Union
from urllib import request
from datetime import datetime, timezone
from .mixin_cli import CLI_Mixin
from .database import Database
from .miscTools import upOut, createDirName, generic_hash, camelCase
from .handleDictionaries import fillDocBeforeCreate, diffDicts
from .miscTools import outputString
from .fixedStringsJson import defaultConfiguration, configurationGUI

class Backend(CLI_Mixin):
  """
  PYTHON BACKEND
  """

  def __init__(self, defaultProjectGroup:str='', **kwargs:int):
    """
    open server and define database

    Args:
        defaultProjectGroup (string): name of configuration / project-group used; if not given, use the one defined by 'defaultProjectGroup' in config file
        **kwargs (dict): additional parameters
          - initViews (bool): initialize views at startup
          - resetDataHierarchy (bool): reset dataHierarchy on database from one on file
    """
    #initialize basic values
    self.hierStack:list[str] = []
    self.alive               = True
    self.cwd:Optional[Path]  = Path('.')
    self.initialize(defaultProjectGroup, **kwargs)


  def initialize(self, defaultProjectGroup:str="", **kwargs:int) -> None:
    """
    initialize or reinitialize server and define database

    Args:
        defaultProjectGroup (string): name of configuration / project-group used; if not given, use the one defined by 'defaultProjectGroup' in config file
        **kwargs (dict): additional parameters
          - initViews (bool): initialize views at startup
          - resetDataHierarchy (bool): reset dataHierarchy on database from one on file
    """
    configFileName = Path.home()/'.pastaELN.json'
    self.configuration = defaultConfiguration
    if configFileName.is_file():
      with open(configFileName,'r', encoding='utf-8') as confFile:
        self.configuration |= json.load(confFile)
    for _, items in configurationGUI.items():
      for k,v in items.items():
        if k not in self.configuration['GUI']:
          self.configuration['GUI'][k] = v[1]
    if self.configuration['version']!=2:
      print('**ERROR Configuration file does not exist or version is not 2')
      return
    if not defaultProjectGroup:
      defaultProjectGroup = self.configuration['defaultProjectGroup']
    if defaultProjectGroup not in self.configuration['projectGroups']:
      raise ValueError('BadConfigurationFileError')
    projectGroup = self.configuration['projectGroups'][defaultProjectGroup]
    if 'user' in projectGroup['local']:
      n,s = projectGroup['local']['user'], projectGroup['local']['password']
    else:
      n,s = upOut(projectGroup['local']['cred'])[0].split(':')
    databaseName = projectGroup['local']['database']
    # directories
    #    self.basePath (root of directory tree) is root of all projects
    #    self.cwd changes during program but is similarly the full path from root
    self.basePath      = Path(projectGroup['local']['path'])
    self.cwd           = Path(projectGroup['local']['path'])
    self.extractorPath = Path(self.configuration['extractorDir'])
    sys.path.append(str(self.extractorPath))  #allow extractors
    # decipher miscellaneous configuration and store
    self.userID   = self.configuration['userID']
    # start database
    self.db = Database(n,s,databaseName, self.configuration, basePath=self.basePath)
    if not hasattr(self.db, 'databaseName'):  #not successful database creation
      return
    if kwargs.get('initViews', False):
      self.db.initDocTypeViews(self.configuration['tableColumnsMax'])
    # internal hierarchy structure
    self.hierStack = []
    self.alive     = True
    return


  def exit(self, deleteDB:bool=False) -> None:
    """
    Shutting down things

    Args:
      deleteDB (bool): remove database
    """
    self.db.exit(deleteDB)
    self.alive     = False
    return


  ######################################################
  ### Change in database
  ######################################################
  def editData(self, doc:dict[str,Any]) -> None:
    """
    Edit data from version 2 information flow by wrapping addData

    Args:
      doc (dict): dict to save
    """
    doc = dict(doc)
    if doc['-branch'][0]['path'] is None:
      self.cwd     = None
    else:
      self.cwd     = self.basePath/doc['-branch'][0]['path']
    self.hierStack = doc['-branch'][0]['stack']+[doc['_id']]
    doc['childNum']= doc['-branch'][0]['child']
    # change content
    self.addData('-edit-', doc)
    # change folder-name in database of all children
    if doc['-type'][0][0]=='x' and self.cwd is not None:
      items = self.db.getView('viewHierarchy/viewPaths',
                              startKey=f'{self.cwd.relative_to(self.basePath).as_posix()}/')
      for item in items:
        t = Thread(target=self.threadParallel, args=(item, doc))
        t.start()
    self.cwd = self.basePath #reset to sensible before continuing
    self.hierStack = []
    return


  def threadParallel(self, item:dict[str,Any], doc:dict[str,Any]) -> None:
    """ internal function to process items and documents in parallel: rename folders on the disk

    Args:
      item (dict): item of current item to process
      doc (dict):  parents documents
    """
    oldPathParts = item['key'].split('/')
    newPathParts = doc['-branch']['path'].split('/')
    newPath = '/'.join(newPathParts+oldPathParts[len(newPathParts):]  )
    if newPath != item['key']:  # for-loop could also be implemented in parallel
      # print(item['id']+'  old='+item['key']+'  branch='+str(item['value'][-1])+\
      #      '  child='+str(item['value'][-3])+'  new='+newPath)
      self.db.updateBranch(item['id'], item['value'][-1], item['value'][-3], path=newPath)
    return


  def addData(self, docType:str, doc:dict[str,Any], hierStack:list[str]=[], localCopy:bool=False,
              forceNewImage:bool=False) -> str:
    """
    Save doc to database, also after edit

    Args:
        docType (string): docType to be stored, subtypes are / separated; or '-edit-'
        doc (dict): to be stored
        hierStack (list): hierStack from external functions
        localCopy (bool): copy a remote file to local version
        forceNewImage (bool): create new image in any case

    Returns:
        str: docID, empty string if failure
    """
    doc['-user']  = self.userID
    childNum     = doc.pop('childNum',None)
    path         = None
    oldPath      = None
    operation    = 'c'  #operation of branch/path
    if docType == '-edit-':
      edit = True
      if '-type' not in doc:
        doc['-type'] = [f'x{len(self.hierStack)}']
      if not hierStack:
        hierStack = self.hierStack
      if '_id' not in doc:
        doc['_id'] = hierStack[-1]
      if len(hierStack)>0 and doc['-type'][0][0]=='x':
        hierStack  = hierStack[:-1]
        oldPath    =  doc['-branch'][0]['path']
      elif '-branch' in doc:
        hierStack   = doc['-branch'][0]['stack']
    else:  #new doc
      edit = False
      doc['-type'] = docType.split('/')
      if len(hierStack) == 0:
        hierStack = self.hierStack
    logging.debug((((
        f'Add/edit data in cwd:{str(self.cwd)} with stack:{str(hierStack)} and name: '
        + doc['-name']) + ' type:') + str(doc['-type']) + ' and edit: ') +
                  str(edit))

    # collect structure-doc and prepare
    if doc['-type'][0][0]=='x' and doc['-type'][0]!='x0' and childNum is None:
      #should not have childnumber in other cases
      thisStack = ' '.join(hierStack)
      view = self.db.getView('viewHierarchy/viewHierarchy', startKey=thisStack) #not faster with cT.getChildren
      childNum = 0
      for item in view:
        if item['value'][1][0]=='x0' or item['value'][1][0][0]!='x':
          continue
        if thisStack == ' '.join(item['key'].split(' ')[:-1]): #remove last item from string
          childNum += 1

    # find path name on local file system; name can be anything
    if self.cwd is not None and '-name' in doc:
      if doc['-type'][0][0]=='x':
        #project, step, task
        if doc['-type'][0]=='x0':
          childNum = 0
        #parentDir
        #  edit: cwd of the project/step/task: remove last directory from cwd (since cwd contains a / at end: remove two)
        #  new: below the current project/step/task
        parentDirectory = self.cwd.parent if edit else self.cwd
        path = parentDirectory/createDirName(doc['-name'],doc['-type'][0],childNum) #update,or create (if new doc, update ignored anyhow)
        operation = 'u'
      else:
        #measurement, sample, procedure
        shasum = ''
        if '://' in doc['-name']:                           #make up name
          if localCopy:
            baseName  = Path(doc['-name']).stem
            extension = Path(doc['-name']).suffix
            path = self.cwd/(camelCase(baseName)+extension)
            request.urlretrieve(doc['-name'], path)
            doc['-name'] = camelCase(baseName)+extension
          else:
            path = Path(doc['-name'])
            try:
              shasum  = generic_hash(path)
            except Exception:
              print('**ERROR bad01: fetch remote content failed. Data not added')
              return ''
        elif doc['-name']!='' and (self.basePath/doc['-name']).is_file():          #file exists
          path = self.basePath/doc['-name']
          doc['-name'] = Path(doc['-name']).name
        elif doc['-name']!='' and (self.cwd/doc['-name']).is_file():               #file exists
          path = self.cwd/doc['-name']
        elif '-branch' in doc:
          if len(doc['-branch'])==1:
            if doc['-branch'][0]['path'] is not None and (self.basePath/doc['-branch'][0]['path']).is_file():
              path = self.basePath/doc['-branch'][0]['path']
          else:
            logging.warning('backend: add document with multiple branches'+str(doc['-branch']) )
        else:                                                                     #make up name
          shasum  = '-'
        if shasum!='-' and path is not None:
          if shasum == '':
            shasum = generic_hash(path, forceFile=True)
          view = self.db.getView('viewIdentify/viewSHAsum',shasum)
          if len(view)==0 or forceNewImage:  #measurement not in database: create doc
            self.useExtractors(path,shasum,doc)  #create image/content
            # All files should appear in database
            # if not 'image' in doc and not 'content' in doc and not 'otherELNName' in doc:  #did not get valuable data: extractor does not exit
            #   return ''
          if len(view)==1:  #measurement is already in database
            doc['_id'] = view[0]['id']
            doc['shasum'] = shasum
            if doc['-type'] == ['-']:  #don't overwrite identified type, without going through extractor
              del doc['-type']
            edit = True
    # assemble branch information
    if childNum is None:
      childNum=9999
    if path is not None and path.is_absolute():
      path = path.relative_to(self.basePath)
    pathStr = None if path is None else path.as_posix()
    show = [True]*(len(hierStack)+1)
    if '-branch' in doc and len(hierStack)+1==len(doc['-branch'][0]['show']):
      show = doc['-branch'][0]['show']
    doc['-branch'] = {'stack':hierStack,'child':childNum,'path':pathStr, 'show':show, 'op':operation}
    if edit:
      #update document
      keysNone = [key for key in doc if doc[key] is None]
      doc = fillDocBeforeCreate(doc, ['--'])  #store None entries and save back since js2py gets equalizes undefined and null
      for key in keysNone:
        doc[key]=None
      doc = self.db.updateDoc(doc, doc['_id'])
    else:
      # add doc to database
      doc = fillDocBeforeCreate(doc, doc['-type'])
      doc = self.db.saveDoc(doc)

    ## adaptation of directory tree, information on disk: documentID is required
    if self.cwd is not None and doc['-type'][0][0]=='x':
      #project, step, task
      path = Path(doc['-branch'][0]['path'])
      if edit and oldPath is not None:
        if not (self.basePath/oldPath).is_dir():
          print(f'**WARNING: addData edit of folder should have oldPath and that should exist:{oldPath}'
                f'\n This can be triggered if user moved the folder.')
          return ''
        (self.basePath/oldPath).rename(self.basePath/path)
      else:
        (self.basePath/path).mkdir(exist_ok=True)   #if exist, create again; moving not necessary since directory moved in changeHierarchy
      with open(self.basePath/path/'.id_pastaELN.json','w', encoding='utf-8') as f:  #local path, update in any case
        f.write(json.dumps(doc))
    return doc['_id']


  ######################################################
  ### Disk directory/folder methods
  ######################################################
  def changeHierarchy(self, docID:Optional[str], dirName:Optional[Path]=None) -> None:
    """
    Change through text hierarchy structure
    change hierarchyStack, change directory, change stored cwd

    Args:
        docID (string): information on how to change
        dirName (string): change into this directory (absolute path given). For if data is moved
    """
    logging.debug('changeHierarchy should only be used in CLI mode')
    if self.cwd is None:
      return
    if docID is None or (docID[0]=='x' and docID[1]!='-'):  #cd ..: none. close 'project', 'task'
      self.hierStack.pop()
      self.cwd = self.cwd.parent
    elif dirName is None:  # existing ID is given: open that
      doc = self.db.getDoc(docID)
      self.cwd = self.basePath/doc['-branch'][0]['path']
      self.hierStack = doc['-branch'][0]['stack']+[docID]
    else:
      self.cwd = dirName
      self.hierStack.append(docID)
    return


  def scanProject(self, progressBar:Any , projID:str, projPath:str='') -> None:
    """ Scan directory tree recursively from project/...
    - find changes on file system and move those changes to DB
    - use .id_pastaELN.json to track changes of directories, aka projects/steps/tasks
    - use shasum to track changes of measurements etc. (one file=one shasum=one entry in DB)
    - create database entries for measurements in directory
    - move/copy/delete allowed as the doc['path'] = list of all copies
      doc['path'] is adopted once changes are observed

    Args:
      progressBar (QProgressBar): gui - qt progress bar
      projID (str): project's docID
      projPath (str): project's path from basePath; if not given, will be determined

    Raises:
      ValueError: could not add new measurement to database
    """
    progressBar.show()
    self.hierStack = [projID]
    if not projPath:
      projPath = self.db.getDoc(projID)['-branch'][0]['path']
    self.cwd = self.basePath/projPath
    rerunScanTree = False
    #prepare lists and start iterating
    inDB_all = self.db.getView('viewHierarchy/viewPaths')
    pathsInDB_x    = [i['key'] for i in inDB_all if i['value'][1][0][0]=='x']  #all structure elements: task, subtasts
    pathsInDB_data = [i['key'] for i in inDB_all if i['value'][1][0][0]!='x']
    filesCountSum = sum(len(files) for (r, d, files) in os.walk(self.cwd))
    filesCount = 0
    for root, dirs, files in os.walk(self.cwd, topdown=True):
      #find parent-document
      self.cwd = Path(root).relative_to(self.basePath)
      if self.cwd.name.startswith('trash_'):
        del dirs
        del files
        continue
      parentIDs = [i for i in inDB_all if i['key']==self.cwd.as_posix()]
      if not parentIDs: #skip newly moved folder, will be scanned upon re-scanning
        continue
      parentID = parentIDs[0]['id']
      parentDoc = self.db.getDoc(parentID)
      hierStack = parentDoc['-branch'][0]['stack']+[parentID]
      # handle directories
      for dirName in dirs[::-1]: #sorted forward in Linux
        if dirName.startswith('.') or dirName.startswith('trash_'):
          continue
        path = (Path(root)/dirName).relative_to(self.basePath).as_posix()
        if path in pathsInDB_x: #path already in database
          pathsInDB_x.remove(path)
          continue
        if (self.basePath/path/'.id_pastaELN.json').is_file(): # update branch: path and stack
          with open(self.basePath/path/'.id_pastaELN.json', 'r', encoding='utf-8') as fIn:
            doc = json.loads(fIn.read())
          if (self.basePath/doc['-branch'][0]['path']).parent.as_posix()  == root and \
               doc['-branch'][0]['stack']==hierStack:
            # special case: user wants to have a different directory name in same folder: then the child-number should not change
            childNum = doc['-branch'][0]['child']
            newPath = path
          else:
            #determine childNumber
            thisStack = ' '.join(hierStack)
            view = self.db.getView('viewHierarchy/viewHierarchy', startKey=thisStack)
            childNum = 0
            for item in view:
              if item['value'][1][0]=='x0' or item['value'][1][0][0]!='x':
                continue
              if thisStack == ' '.join(item['key'].split(' ')[:-1]): #remove last item from string
                childNum += 1
            newPath = '/'.join(path.split('/')[:-1])+'/'+createDirName(doc['-name'],doc['-type'][0],childNum) #update,or create (if new doc, update ignored anyhow)
            if (self.basePath/newPath).exists():                     #can be either file or directory
              print("**ERROR new path should not exist",newPath)
            else:
              (self.basePath/path).rename(self.basePath/newPath)
          self.db.updateBranch(doc['_id'], 0, childNum, hierStack, newPath)
        else:
          currentID = self.addData(f'x{len(hierStack)}', {'-name': dirName}, hierStack)
          newDir = self.basePath/self.db.getDoc(currentID)['-branch'][0]['path']
          (newDir/'.id_pastaELN.json').rename(self.basePath/root/dirName/'.id_pastaELN.json') #move index file into old folder
          newDir.rmdir()                     #remove created path
          (self.basePath/root/dirName).rename(newDir) #move old to new path
        rerunScanTree = True
      # handle files
      for fileName in files:
        filesCount += 1
        progressBar.setValue(int(100*filesCount/filesCountSum))
        if fileName.startswith('.') or fileName.startswith('trash_') or '_PastaExport' in fileName:
          continue
        path = (Path(root).relative_to(self.basePath) /fileName).as_posix()
        if path in pathsInDB_data:
          logging.info('Scan: file already in DB: %s',path)
          pathsInDB_data.remove(path)
        else:
          logging.info('Scan: add file to DB: %s',path)
          self.addData('-', {'-name':path}, hierStack)
    #finish method
    self.cwd = self.basePath/projPath
    orphans = [
        i for i in pathsInDB_data
        if i.startswith(f'{self.cwd.relative_to(self.basePath).as_posix()}/')
    ]
    logging.info('Scan: these files are on DB but not hard disk\n'+'\n  '.join(orphans))
    orphanDirs = [
        i for i in pathsInDB_x
        if i.startswith(f'{self.cwd.relative_to(self.basePath).as_posix()}/')
        and i != projPath
    ]
    logging.info('Scan: these directories are on DB but not hard disk\n'+'\n  '.join(orphanDirs))
    for orphan in orphans+orphanDirs:
      docID = [i for i in inDB_all if i['key']==orphan][0]['id']
      doc   = dict(self.db.getDoc(docID))
      change = None
      for branch in doc['-branch']:
        if branch['path']==orphan:
          change = {'-branch': {'op':'d', 'oldpath':branch['path'], 'path':branch['path'], 'stack':branch['stack'] }}
          break
      if change is None:
        logging.warning('Tried to remove orphan in database but could not; that can happen if user renames folder: %s', orphan)
      else:
        self.db.updateDoc(change, docID)
    #reset to initial values
    self.hierStack = []
    self.cwd = Path(self.basePath)
    if rerunScanTree:
      self.scanProject(progressBar, projID, projPath)
    progressBar.hide()
    return


  def useExtractors(self, filePath:Path, shasum:str, doc:dict[str,Any]) -> None:
    """
    get measurements from datafile: central distribution point
    - max image size defined here

    Args:
        filePath (Path): path to file
        shasum (string): shasum (git-style hash) to store in database (not used here)
        doc (dict): pass known data/measurement type, can be used to create image; This doc is altered
    """
    extension = filePath.suffix[1:]  #cut off initial . of .jpg
    if str(filePath).startswith('http'):
      absFilePath = Path(tempfile.gettempdir())/filePath.name
      with request.urlopen(filePath.as_posix().replace(':/','://'), timeout=60) as urlRequest:
        with open(absFilePath, 'wb') as f:
          try:
            f.write(urlRequest.read())
          except Exception:
            print("Error saving downloaded file to temporary disk")
    else:
      if filePath.is_absolute():
        filePath = filePath.relative_to(self.basePath)
      absFilePath = self.basePath/filePath
    pyFile = f'extractor_{extension.lower()}.py'
    pyPath = self.extractorPath/pyFile
    success = False
    if pyPath.is_file():
      success = True
      # import module and use to get data
      os.environ['QT_API'] = 'pyside2'
      import matplotlib.pyplot as plt  #IMPORTANT: NO PYPLOT OUTSIDE THIS QT_API BLOCK
      plt.clf()
      try:
        module  = importlib.import_module(pyFile[:-3])
        content = module.use(absFilePath, '/'.join(doc['-type']) )
      except Exception:
        logging.error(f'ERROR with extractor {pyFile}' + '\n' + traceback.format_exc())
        success = False
      os.environ['QT_API'] = 'pyside6'
      #combine into document
      if success:
        doc |= content
        for meta in ['metaVendor','metaUser']:
          if meta not in doc:
            doc[meta] = {}
          for item in doc[meta]:
            if isinstance(doc[meta][item], tuple):
              doc[meta][item] = list(doc[meta][item])
            try:
              _ = json.dumps(doc[meta][item])
            except:
              doc[meta][item] = str(doc[meta][item])
              print('**Warning -> stringified  ',meta, item)
        if doc['-type'][0] in [doc['recipe'].split('/')[0], '-']:
          doc['-type']     = doc['recipe'].split('/')
        else:
          #user has strange wish: trust him/her
          logging.info('user has strange wish: trust him/her: '+'/'.join(doc['-type'])+'  '+doc['recipe'])
        del doc['recipe']
        if 'fileExtension' not in doc['metaVendor']:
          doc['metaVendor']['fileExtension'] = extension.lower()
        if 'links' in doc and len(doc['links'])==0:
          del doc['links']
    if not success:
      print('  **Warning, issue with extractor', pyFile)
      logging.warning('Issue with extractor %s', pyFile)
      doc['metaUser'] = {'filename':absFilePath.name, 'extension':absFilePath.suffix,
        'filesize':absFilePath.stat().st_size,
        'created at':datetime.fromtimestamp(absFilePath.stat().st_ctime, tz=timezone.utc).isoformat(),
        'modified at':datetime.fromtimestamp(absFilePath.stat().st_mtime, tz=timezone.utc).isoformat()}
    doc['shasum']=shasum  #essential for logic, always save, unlike image
    return


  def testExtractor(self, filePath:Union[Path,str], extractorPath:Optional[Path]=None, recipe:str='',
                    outputStyle:str='text', saveFig:str='') -> str:
    """
    Args:
      filePath (Path, str): path to the file to be tested
      extractorPath (Path, None): path to the directory with extractors
      recipe (str): recipe in / separated
      outputStyle (str): report in ['print','text','html'] including show images
      saveFig (str): save figure to...; if given stop testing after generating image

    Returns:
      str: short summary or long report
    """
    import base64
    from io import BytesIO
    from PIL import Image
    import cairosvg
    os.environ['QT_API'] = 'pyside2'
    import matplotlib.pyplot as plt
    import matplotlib.axes as mpaxes

    htmlStr= 'Please visit <a href="https://pasta-eln.github.io/pasta-eln/extractors.html#'
    success = True
    if isinstance(filePath, str):
      filePath = Path(filePath)
    if filePath.as_posix().startswith('http'):
      tempFilePath = Path(tempfile.gettempdir())/filePath.name
      request.urlretrieve(filePath.as_posix().replace(':/','://'), tempFilePath)
      filePath = tempFilePath
    report = outputString(outputStyle, 'h2', 'Report on extractor test')
    report += outputString(outputStyle, 'info', f'check file: {str(filePath)}')
    extension = filePath.suffix[1:]
    pyFile = f'extractor_{extension.lower()}.py'
    if extractorPath is None:
      extractorPath = self.extractorPath
    #start testing
    if (extractorPath/pyFile).is_file():
      report += outputString(outputStyle, 'info', f'use extractor: {str(extractorPath / pyFile)}')
    else:
      success = False
      report += outputString(outputStyle, 'error', f'No fitting extractor found:{pyFile}')
    if success:
      try:
        module  = importlib.import_module(pyFile[:-3])
        plt.clf()
        content = module.use(filePath, recipe, saveFig or None)
        if saveFig:
          return report
      except Exception:
        success = False
        report += outputString(outputStyle, 'error', 'Python error in extractor')
        report += outputString(outputStyle, 'error', f'{htmlStr}python-error">website</a>')
        report += outputString(outputStyle, 'error', traceback.format_exc(limit=3))
    if success:
      if 'recipe' in content:
        possibleDocTypes = [i for i in self.db.dataLabels.keys() if i[0]!='x']
        matches = [i for i in possibleDocTypes if content['recipe'].startswith(i)]
        if matches or content['recipe'] == '' or content['recipe'] == '-':
          report += outputString(outputStyle, 'info', 'Recipe is good: '+content['recipe'])
          size = len(str(content))
          report += outputString(outputStyle, 'info', f'Entire extracted size: {size // 1024}kB')
        else:
          report += outputString(outputStyle, 'error', 'Recipe does not follow doctype in dataHierarchy.')
      else:
        report += outputString(outputStyle,'error','Recipe not included in extractor.')
    if success:
      try:
        _ = json.dumps(content)
      except Exception:
        report += outputString(outputStyle,'error','Extractor reply not json dumpable.')
    if success:
      try:
        _ = json.dumps(content['metaVendor'])
        report += outputString(outputStyle,'info','Number of vendor entries: '+str(len(content['metaVendor'])))
      except Exception:
        # possible cause of failure: make sure that no int64 but normal int
        success = False
        report += outputString(outputStyle,'error', 'Some json format does not fit in metaVendor')
        report += outputString(outputStyle, 'error', f'{htmlStr}metadata-error">website</a>')
        #iterate keys
        for key in content['metaVendor']:
          try:
            _ = json.dumps(content['metaVendor'][key])
          except Exception:
            report += outputString(outputStyle,'error',f'FAIL {key}:'+str(content['metaVendor'][key])+' type:')+str(type(content['metaVendor'][key]))
    if success:
      try:
        _ = json.dumps(content['metaUser'])
        report += 'Number of user entries: '+str(len(content['metaUser']))+'<br>'
      except Exception:
        report += outputString(outputStyle,'error', 'Some json format does not fit in metaUser')
        report += outputString(outputStyle, 'error', f'{htmlStr}metadata-error">website</a>')
        #iterate keys
        for key in content['metaUser']:
          try:
            _ = json.dumps(content['metaUser'][key])
          except Exception:
            report += outputString(outputStyle,'error',f'FAIL {key}:'+str(content['metaUser'][key])+' type:') + str(type(content['metaUser'][key]))
      if 'image' not in content:
        success = False
        report += outputString(outputStyle,'error','Image does not exist')
    if success and isinstance(content['image'],Image.Image):
      success = False
      report += outputString(outputStyle,'error','Image is a PIL image: not a base64 string')
      report += outputString(outputStyle, 'error', f'{htmlStr}pillow-image">website</a>')
        # print('Encode image via the following: pay attention to jpg/png which is encoded twice\n```')
        # print('from io import BytesIO')
        # print('figfile = BytesIO()')
        # print('image.save(figfile, format="PNG")')
        # print('imageData = base64.b64encode(figfile.getvalue()).decode()')
        # print('image = "data:image/jpg;base64," + imageData')
    if success and isinstance(content['image'], mpaxes._axes.Axes): # pylint: disable=protected-access
      success = False
      report += outputString(outputStyle,'error','Image is a matplotlib axis: not a base64 string')
      report += outputString(outputStyle, 'error', f'{htmlStr}matplotlib">website</a>')
        # print('**Warning: image is a matplotlib axis: not a svg string')
        # print('  figfile = StringIO()')
        # print('plt.savefig(figfile, format="svg")')
        # print('image = figfile.getvalue()')
    if success and isinstance(content['image'], str):#show content
      size = len(content['image'])
      report += outputString(outputStyle, 'info', f'Image size {str(size // 1024)}kB')
      if outputStyle!='text':
        report += outputString(outputStyle,'h2','Additional window shows the image')
      if len(content['image'])>20:
        if content['image'].startswith('data:image/'):
          #png or jpg encoded base64
          extension = content['image'][11:14]
          img = base64.b64decode(content['image'][22:])
        else:
          #svg data
          img = cairosvg.svg2png(bytestring=content['image'].encode())
        i = BytesIO(img)
        image = Image.open(i)
        if outputStyle!='text':
          image.show()
      del content['image']
    if outputStyle=='print':
      print('Identified metadata',content)
    os.environ['QT_API'] = 'pyside6'
    return report


  ######################################################
  ### Wrapper for database functions
  ######################################################
  def replicateDB(self, progressBar:Any, removeAtStart:bool=False) -> str:
    """
    Replicate local database to remote database

    Args:
        removeAtStart (bool): remove remote DB before starting new
        progressBar (QProgressBar): gui - qt progress bar

    Returns:
        str: replication report
    """
    defaultProjectGroup = self.configuration['defaultProjectGroup']
    remoteConf = self.configuration['projectGroups'][defaultProjectGroup]['remote']
    if not remoteConf: #empty entry: fails
      print("**ERROR brp01: You tried to replicate although, remote is not defined")
      return 'ERROR'
    if 'cred' in remoteConf:
      remoteConf['user'], remoteConf['password'] = upOut(remoteConf['cred'])[0].split(':')
    return self.db.replicateDB(remoteConf, progressBar, removeAtStart)


  def checkDB(self, outputStyle:str='text', repair:bool=False, minimal:bool=False) -> str:
    """
    Wrapper of check database for consistencies by iterating through all documents

    Args:
        outputStyle (str): output using a given style: see outputString
        repair (bool): repair database
        minimal (bool): true=only show warnings and errors; else=also show information

    Returns:
        string: output incl. \n
    """
    ### check database itself for consistency
    output = self.db.checkDB(outputStyle=outputStyle, repair=repair, minimal=minimal)
    if repair:
      self.db.initGeneralViews()
    ### compare with file system
    output += outputString(outputStyle,'h2','File status')
    viewProjects   = self.db.getView('viewDocType/x0All')
    inDB_all = self.db.getView('viewHierarchy/viewPathsAll')
    pathsInDB_data = [i['key'] for i in inDB_all if i['value'][1][0][0]!='x']
    pathsInDB_folder = [i['key'] for i in inDB_all if i['value'][1][0][0]=='x']
    count = 0
    for projI in viewProjects:
      projDoc = self.db.getDoc(projI['id'])
      if len(projDoc['-branch'])==0:
        output += outputString(outputStyle,'error','project view got screwed up')
        continue
      for root, dirs, files in os.walk(self.basePath/projDoc['-branch'][0]['path']):
        if Path(root).name[0]=='.' or Path(root).name.startswith('trash_'):
          del dirs
          continue
        for fileName in files:
          if fileName.startswith('.') or fileName.startswith('trash_') or '_PastaExport' in fileName:
            continue
          path = (Path(root).relative_to(self.basePath) /fileName).as_posix()
          if path not in pathsInDB_data:
            output += outputString(outputStyle, 'error', f'File on harddisk but not DB: {path}')
            count += 1
          else:
            pathsInDB_data.remove(path)
        for dirName in dirs:
          if dirName.startswith('.') or dirName.startswith('trash_'):
            continue
          path = (Path(root).relative_to(self.basePath) /dirName).as_posix()
          if path not in pathsInDB_folder:
            output += outputString(outputStyle, 'error', f'Directory on harddisk but not DB:{path}')
            count += 1
          else:
            pathsInDB_folder.remove(path)
            if (self.basePath/root/dirName/'.id_pastaELN.json').is_file():
              with open(self.basePath/root/dirName/'.id_pastaELN.json','r',encoding='utf-8') as fIn:
                docDisk = json.loads(fIn.read())
                listDocs = self.db.getView('viewHierarchy/viewPathsAll', preciseKey=path)
                if len(listDocs)!=1:
                  output += outputString(outputStyle, 'error', f'Path of folder is non-unique: {path}')
                docDB   = self.db.getDoc(listDocs[0]['id'])
                difference = diffDicts(docDisk,docDB)
                if len(difference)>1:
                  output += outputString(outputStyle,'error','disk(1) and db(2) content do not match:'+docDisk['_id'])
                  output += outputString(outputStyle,'error',difference)
                  if repair:
                    with open(self.basePath/root/dirName/'.id_pastaELN.json','w',encoding='utf-8') as fOut:
                      json.dump(docDB, fOut)
                  # #use only for resetting the content in the .id_pastaELN.json
            else:
              output += outputString(outputStyle, 'error', f'Folder has no .id_pastaELN.json:{path}')
              count += 1
              # if True:  #use only for resetting the content in the .id_pastaELN.json
              #   with open(self.basePath/root/dirName/'.id_pastaELN.json','w',encoding='utf-8') as fOut:
              #     docDB   = self.db.getDoc( docDisk['_id'] )
              #     json.dump(docDB, fOut)
    output += outputString(outputStyle, 'info', f'Number of files on disk that are not in database {str(count)}')
    orphans = [i for i in pathsInDB_data   if not (self.basePath/i).exists() and ":/" not in i]  #paths can be files or directories
    orphans+= [i for i in pathsInDB_folder if not (self.basePath/i).exists() ]
    if orphans:
      output += outputString(outputStyle,'error','bch01: These files of database not on filesystem:\n- '+'\n- '.join(orphans))
    output += outputString(outputStyle,'h2','File summary')
    if outputStyle == 'text':
      output += "Success\n" if not orphans and count==0 else "Failure\n"
    return output



""" Represents an error in the database. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: database_error.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

class DatabaseError(Exception):
  """
  Represents an error in the database.

  Explanation:
      This class represents a database error and inherits from the Exception class.
      It provides a way to construct a DatabaseError instance with the provided message and detailed_errors.

  Args:
      message (str): The error message.
      detailed_errors (dict[str, str], optional): Additional detailed errors.
      Defaults to None.

  Returns:
      None
  """

  def __init__(self,
               message: str,
               detailed_errors: dict[str, str] | None = None):
    """
    Constructs a DatabaseError instance.

    Explanation:
        This method initializes a DatabaseError instance with the provided message and detailed_errors.
        It inherits from the Exception class.

    Args:
        message (str): The error message.
        detailed_errors (dict[str, str], optional): Additional detailed errors.
        Defaults to None.

    Returns:
        None
    """
    super().__init__(message)
    self.message = message
    self.detailed_errors = detailed_errors



""" Represents a task worker_thread extension. """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2024
#
#  Author: Jithu Murugan
#  Filename: task_thread_extension.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.

import logging

from PySide6 import QtCore
from PySide6.QtCore import QObject

from pasta_eln.dataverse.generic_task_object import GenericTaskObject


class TaskThreadExtension(QObject):
  """
  Represents a task worker_thread extension.

  Explanation:
      This class provides an extension for running a task in a separate worker_thread.
      It initializes the task and worker_thread, and provides a method to quit the task and clean up resources.

  Args:
      task (GenericTaskObject): The task to be executed in the separate worker_thread.

  Returns:
      None
  """

  def __init__(self, task: GenericTaskObject) -> None:
    """
    Initializes the TaskThreadExtension instance.

    Explanation:
        This method initializes the TaskThreadExtension instance by setting the task and worker_thread attributes.
        It moves the task to the worker_thread and starts the worker_thread.

    Args:
        task (GenericTaskObject): The task object to be executed in the separate worker_thread.

    Returns:
        None
    """
    super().__init__()
    self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
    self.task: GenericTaskObject = task
    self.worker_thread: QtCore.QThread = QtCore.QThread()
    self.task.moveToThread(self.worker_thread)
    self.worker_thread.start()

  def quit(self) -> None:
    """
    Quits the task and cleans up resources.

    Explanation:
        This method emits the cancel signal of the task, performs cleanup, and signals the worker_thread to quit.

    Args:
        self: The TaskThreadExtension instance.

    """
    self.logger.info("Quitting task thread extension, Task id: %s", self.task.id)
    self.task.cancel.emit()
    self.task.cleanup()
    self.worker_thread.quit()



# Form implementation generated from reading ui file 'data_hierarchy_editor_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PySide6 import QtCore, QtGui, QtWidgets


class Ui_DataHierarchyEditorDialogBase(object):
  def setupUi(self, DataHierarchyEditorDialogBase):
    DataHierarchyEditorDialogBase.setObjectName("DataHierarchyEditorDialogBase")
    DataHierarchyEditorDialogBase.resize(1271, 845)
    DataHierarchyEditorDialogBase.setToolTip("")
    self.gridLayout = QtWidgets.QGridLayout(DataHierarchyEditorDialogBase)
    self.gridLayout.setContentsMargins(10, 10, 10, 10)
    self.gridLayout.setObjectName("gridLayout")
    self.mainWidget = QtWidgets.QWidget(parent=DataHierarchyEditorDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.mainWidget.sizePolicy().hasHeightForWidth())
    self.mainWidget.setSizePolicy(sizePolicy)
    self.mainWidget.setObjectName("mainWidget")
    self.gridLayout_2 = QtWidgets.QGridLayout(self.mainWidget)
    self.gridLayout_2.setObjectName("gridLayout_2")
    self.mainGridLayout = QtWidgets.QGridLayout()
    self.mainGridLayout.setContentsMargins(30, -1, 30, -1)
    self.mainGridLayout.setObjectName("mainGridLayout")
    self.typeMetadataTableView = QtWidgets.QTableView(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeMetadataTableView.sizePolicy().hasHeightForWidth())
    self.typeMetadataTableView.setSizePolicy(sizePolicy)
    self.typeMetadataTableView.setSortingEnabled(False)
    self.typeMetadataTableView.setObjectName("typeMetadataTableView")
    self.typeMetadataTableView.horizontalHeader().setCascadingSectionResizes(True)
    self.typeMetadataTableView.horizontalHeader().setSortIndicatorShown(False)
    self.typeMetadataTableView.horizontalHeader().setStretchLastSection(False)
    self.typeMetadataTableView.verticalHeader().setCascadingSectionResizes(True)
    self.typeMetadataTableView.verticalHeader().setSortIndicatorShown(False)
    self.typeMetadataTableView.verticalHeader().setStretchLastSection(False)
    self.mainGridLayout.addWidget(self.typeMetadataTableView, 6, 0, 1, 1)
    self.attachmentsHeaderHorizontalLayout = QtWidgets.QHBoxLayout()
    self.attachmentsHeaderHorizontalLayout.setContentsMargins(0, 5, 0, 5)
    self.attachmentsHeaderHorizontalLayout.setObjectName("attachmentsHeaderHorizontalLayout")
    self.attachmentsShowHidePushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.attachmentsShowHidePushButton.sizePolicy().hasHeightForWidth())
    self.attachmentsShowHidePushButton.setSizePolicy(sizePolicy)
    self.attachmentsShowHidePushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.attachmentsShowHidePushButton.setObjectName("attachmentsShowHidePushButton")
    self.attachmentsHeaderHorizontalLayout.addWidget(self.attachmentsShowHidePushButton)
    spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.attachmentsHeaderHorizontalLayout.addItem(spacerItem)
    self.mainGridLayout.addLayout(self.attachmentsHeaderHorizontalLayout, 10, 0, 1, 1)
    self.metadataGroupHorizontalLayout = QtWidgets.QHBoxLayout()
    self.metadataGroupHorizontalLayout.setContentsMargins(0, 5, 0, 5)
    self.metadataGroupHorizontalLayout.setObjectName("metadataGroupHorizontalLayout")
    self.metadataGroupLabel = QtWidgets.QLabel(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.metadataGroupLabel.sizePolicy().hasHeightForWidth())
    self.metadataGroupLabel.setSizePolicy(sizePolicy)
    self.metadataGroupLabel.setMinimumSize(QtCore.QSize(130, 0))
    self.metadataGroupLabel.setObjectName("metadataGroupLabel")
    self.metadataGroupHorizontalLayout.addWidget(self.metadataGroupLabel)
    self.metadataGroupComboBox = QtWidgets.QComboBox(parent=self.mainWidget)
    self.metadataGroupComboBox.setEnabled(True)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.metadataGroupComboBox.sizePolicy().hasHeightForWidth())
    self.metadataGroupComboBox.setSizePolicy(sizePolicy)
    self.metadataGroupComboBox.setMinimumSize(QtCore.QSize(200, 0))
    self.metadataGroupComboBox.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToContents)
    self.metadataGroupComboBox.setObjectName("metadataGroupComboBox")
    self.metadataGroupHorizontalLayout.addWidget(self.metadataGroupComboBox)
    self.addMetadataGroupLineEdit = QtWidgets.QLineEdit(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.addMetadataGroupLineEdit.sizePolicy().hasHeightForWidth())
    self.addMetadataGroupLineEdit.setSizePolicy(sizePolicy)
    self.addMetadataGroupLineEdit.setMinimumSize(QtCore.QSize(0, 0))
    self.addMetadataGroupLineEdit.setClearButtonEnabled(True)
    self.addMetadataGroupLineEdit.setObjectName("addMetadataGroupLineEdit")
    self.metadataGroupHorizontalLayout.addWidget(self.addMetadataGroupLineEdit)
    self.addMetadataGroupPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.addMetadataGroupPushButton.sizePolicy().hasHeightForWidth())
    self.addMetadataGroupPushButton.setSizePolicy(sizePolicy)
    self.addMetadataGroupPushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.addMetadataGroupPushButton.setStatusTip("")
    self.addMetadataGroupPushButton.setObjectName("addMetadataGroupPushButton")
    self.metadataGroupHorizontalLayout.addWidget(self.addMetadataGroupPushButton)
    self.deleteMetadataGroupPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.deleteMetadataGroupPushButton.sizePolicy().hasHeightForWidth())
    self.deleteMetadataGroupPushButton.setSizePolicy(sizePolicy)
    self.deleteMetadataGroupPushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.deleteMetadataGroupPushButton.setObjectName("deleteMetadataGroupPushButton")
    self.metadataGroupHorizontalLayout.addWidget(self.deleteMetadataGroupPushButton)
    self.mainGridLayout.addLayout(self.metadataGroupHorizontalLayout, 3, 0, 1, 1)
    self.datatypeHorizontalLayout = QtWidgets.QHBoxLayout()
    self.datatypeHorizontalLayout.setContentsMargins(0, 5, 0, 5)
    self.datatypeHorizontalLayout.setObjectName("datatypeHorizontalLayout")
    self.typeLabel = QtWidgets.QLabel(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Preferred)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeLabel.sizePolicy().hasHeightForWidth())
    self.typeLabel.setSizePolicy(sizePolicy)
    self.typeLabel.setMinimumSize(QtCore.QSize(130, 0))
    self.typeLabel.setObjectName("typeLabel")
    self.datatypeHorizontalLayout.addWidget(self.typeLabel)
    self.typeComboBox = QtWidgets.QComboBox(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeComboBox.sizePolicy().hasHeightForWidth())
    self.typeComboBox.setSizePolicy(sizePolicy)
    self.typeComboBox.setMinimumSize(QtCore.QSize(200, 0))
    self.typeComboBox.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToContents)
    self.typeComboBox.setObjectName("typeComboBox")
    self.datatypeHorizontalLayout.addWidget(self.typeComboBox)
    self.typeDisplayedTitleLineEdit = QtWidgets.QLineEdit(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeDisplayedTitleLineEdit.sizePolicy().hasHeightForWidth())
    self.typeDisplayedTitleLineEdit.setSizePolicy(sizePolicy)
    self.typeDisplayedTitleLineEdit.setMinimumSize(QtCore.QSize(0, 0))
    self.typeDisplayedTitleLineEdit.setText("")
    self.typeDisplayedTitleLineEdit.setClearButtonEnabled(True)
    self.typeDisplayedTitleLineEdit.setObjectName("typeDisplayedTitleLineEdit")
    self.datatypeHorizontalLayout.addWidget(self.typeDisplayedTitleLineEdit)
    self.typeIriLineEdit = QtWidgets.QLineEdit(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeIriLineEdit.sizePolicy().hasHeightForWidth())
    self.typeIriLineEdit.setSizePolicy(sizePolicy)
    self.typeIriLineEdit.setClearButtonEnabled(True)
    self.typeIriLineEdit.setObjectName("typeIriLineEdit")
    self.datatypeHorizontalLayout.addWidget(self.typeIriLineEdit)
    self.addTypePushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.addTypePushButton.sizePolicy().hasHeightForWidth())
    self.addTypePushButton.setSizePolicy(sizePolicy)
    self.addTypePushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.addTypePushButton.setObjectName("addTypePushButton")
    self.datatypeHorizontalLayout.addWidget(self.addTypePushButton)
    self.deleteTypePushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.deleteTypePushButton.sizePolicy().hasHeightForWidth())
    self.deleteTypePushButton.setSizePolicy(sizePolicy)
    self.deleteTypePushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.deleteTypePushButton.setObjectName("deleteTypePushButton")
    self.datatypeHorizontalLayout.addWidget(self.deleteTypePushButton)
    self.mainGridLayout.addLayout(self.datatypeHorizontalLayout, 1, 0, 1, 1)
    self.metadataTableButtonHorizontalLayout = QtWidgets.QHBoxLayout()
    self.metadataTableButtonHorizontalLayout.setContentsMargins(0, 5, 0, 5)
    self.metadataTableButtonHorizontalLayout.setObjectName("metadataTableButtonHorizontalLayout")
    self.addMetadataRowPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.addMetadataRowPushButton.sizePolicy().hasHeightForWidth())
    self.addMetadataRowPushButton.setSizePolicy(sizePolicy)
    self.addMetadataRowPushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.addMetadataRowPushButton.setObjectName("addMetadataRowPushButton")
    self.metadataTableButtonHorizontalLayout.addWidget(self.addMetadataRowPushButton)
    spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.metadataTableButtonHorizontalLayout.addItem(spacerItem1)
    self.mainGridLayout.addLayout(self.metadataTableButtonHorizontalLayout, 8, 0, 1, 1)
    self.attachmentTableButtonsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.attachmentTableButtonsHorizontalLayout.setContentsMargins(0, 5, 0, 5)
    self.attachmentTableButtonsHorizontalLayout.setObjectName("attachmentTableButtonsHorizontalLayout")
    self.addAttachmentPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    self.addAttachmentPushButton.setMinimumSize(QtCore.QSize(200, 0))
    self.addAttachmentPushButton.setObjectName("addAttachmentPushButton")
    self.attachmentTableButtonsHorizontalLayout.addWidget(self.addAttachmentPushButton)
    spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.attachmentTableButtonsHorizontalLayout.addItem(spacerItem2)
    self.mainGridLayout.addLayout(self.attachmentTableButtonsHorizontalLayout, 13, 0, 1, 1)
    self.metadataTableHeaderHorizontalLayout = QtWidgets.QHBoxLayout()
    self.metadataTableHeaderHorizontalLayout.setContentsMargins(-1, 5, -1, 5)
    self.metadataTableHeaderHorizontalLayout.setObjectName("metadataTableHeaderHorizontalLayout")
    self.metadataTableHeaderLabel = QtWidgets.QLabel(parent=self.mainWidget)
    font = QtGui.QFont()
    font.setBold(True)
    self.metadataTableHeaderLabel.setFont(font)
    self.metadataTableHeaderLabel.setObjectName("metadataTableHeaderLabel")
    self.metadataTableHeaderHorizontalLayout.addWidget(self.metadataTableHeaderLabel)
    self.mainGridLayout.addLayout(self.metadataTableHeaderHorizontalLayout, 4, 0, 1, 1)
    self.typeAttachmentsTableView = QtWidgets.QTableView(parent=self.mainWidget)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.typeAttachmentsTableView.sizePolicy().hasHeightForWidth())
    self.typeAttachmentsTableView.setSizePolicy(sizePolicy)
    self.typeAttachmentsTableView.setObjectName("typeAttachmentsTableView")
    self.typeAttachmentsTableView.horizontalHeader().setStretchLastSection(False)
    self.mainGridLayout.addWidget(self.typeAttachmentsTableView, 12, 0, 1, 1)
    self.headerHorizontalLayout = QtWidgets.QHBoxLayout()
    self.headerHorizontalLayout.setContentsMargins(0, 20, 0, 20)
    self.headerHorizontalLayout.setObjectName("headerHorizontalLayout")
    self.headerLabel = QtWidgets.QLabel(parent=self.mainWidget)
    font = QtGui.QFont()
    font.setPointSize(14)
    font.setBold(True)
    self.headerLabel.setFont(font)
    self.headerLabel.setObjectName("headerLabel")
    self.headerHorizontalLayout.addWidget(self.headerLabel)
    spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.headerHorizontalLayout.addItem(spacerItem3)
    self.saveDataHierarchyPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    self.saveDataHierarchyPushButton.setObjectName("saveDataHierarchyPushButton")
    self.headerHorizontalLayout.addWidget(self.saveDataHierarchyPushButton)
    self.helpPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    self.helpPushButton.setObjectName("helpPushButton")
    self.headerHorizontalLayout.addWidget(self.helpPushButton)
    self.cancelPushButton = QtWidgets.QPushButton(parent=self.mainWidget)
    self.cancelPushButton.setObjectName("cancelPushButton")
    self.headerHorizontalLayout.addWidget(self.cancelPushButton)
    self.mainGridLayout.addLayout(self.headerHorizontalLayout, 0, 0, 1, 1)
    self.gridLayout_2.addLayout(self.mainGridLayout, 0, 0, 1, 1)
    self.gridLayout.addWidget(self.mainWidget, 0, 1, 1, 1)

    self.retranslateUi(DataHierarchyEditorDialogBase)
    QtCore.QMetaObject.connectSlotsByName(DataHierarchyEditorDialogBase)

  def retranslateUi(self, DataHierarchyEditorDialogBase):
    _translate = QtCore.QCoreApplication.translate
    DataHierarchyEditorDialogBase.setWindowTitle(_translate("DataHierarchyEditorDialogBase", "Data Hierarchy Editor"))
    self.typeMetadataTableView.setToolTip(_translate("DataHierarchyEditorDialogBase", "Table for all metadata associated with the Data Type. Add \"comment\" or \"content\" for editable text fields, \"image\" for image support, or enter another Data Type to enable links."))
    self.attachmentsShowHidePushButton.setText(_translate("DataHierarchyEditorDialogBase", "Show/Hide Attachments"))
    self.metadataGroupLabel.setText(_translate("DataHierarchyEditorDialogBase", "Metadata Group"))
    self.metadataGroupComboBox.setToolTip(_translate("DataHierarchyEditorDialogBase", "Select the group of metadata to be listed below in the table"))
    self.addMetadataGroupLineEdit.setToolTip(_translate("DataHierarchyEditorDialogBase", "Enter the new group to be added to the data type"))
    self.addMetadataGroupLineEdit.setPlaceholderText(_translate("DataHierarchyEditorDialogBase", "Enter the new group to be added"))
    self.addMetadataGroupPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Add a new group of metadata to the data type, table below will be reset to empty list!"))
    self.addMetadataGroupPushButton.setText(_translate("DataHierarchyEditorDialogBase", "+ Add"))
    self.deleteMetadataGroupPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Delete the selected group in the group combobox"))
    self.deleteMetadataGroupPushButton.setText(_translate("DataHierarchyEditorDialogBase", "- Delete"))
    self.typeLabel.setText(_translate("DataHierarchyEditorDialogBase", "Data Type"))
    self.typeComboBox.setToolTip(_translate("DataHierarchyEditorDialogBase", "Select the type from the loaded data hierarchy types"))
    self.typeDisplayedTitleLineEdit.setToolTip(_translate("DataHierarchyEditorDialogBase", "Modify the displayed title property of the type"))
    self.typeDisplayedTitleLineEdit.setPlaceholderText(_translate("DataHierarchyEditorDialogBase", "Modify the type displayed title here"))
    self.typeIriLineEdit.setToolTip(_translate("DataHierarchyEditorDialogBase", "Enter the link/iri to be associated with this data-type"))
    self.typeIriLineEdit.setPlaceholderText(_translate("DataHierarchyEditorDialogBase", "Enter the IRI for the type"))
    self.addTypePushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Add a new type (structural or normal type) to the data hierarchy data set."))
    self.addTypePushButton.setText(_translate("DataHierarchyEditorDialogBase", "+ Add"))
    self.deleteTypePushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Delete the type with the full metadata and attachments completely"))
    self.deleteTypePushButton.setText(_translate("DataHierarchyEditorDialogBase", "- Delete"))
    self.addMetadataRowPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Add a new metadata row to the above table with empty values"))
    self.addMetadataRowPushButton.setText(_translate("DataHierarchyEditorDialogBase", "+ Add Metadata"))
    self.addAttachmentPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Add a new attachment row to the above table with empty values"))
    self.addAttachmentPushButton.setText(_translate("DataHierarchyEditorDialogBase", "+ Add Attachment"))
    self.metadataTableHeaderLabel.setText(_translate("DataHierarchyEditorDialogBase", "Metadata"))
    self.typeAttachmentsTableView.setToolTip(_translate("DataHierarchyEditorDialogBase", "Table which displays the attachments for the above selected data type"))
    self.headerLabel.setText(_translate("DataHierarchyEditorDialogBase", "Edit the data hierarchy for the PASTA-ELN projects"))
    self.saveDataHierarchyPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Save loaded data hierarchy in local database"))
    self.saveDataHierarchyPushButton.setText(_translate("DataHierarchyEditorDialogBase", "Save"))
    self.helpPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Navigate to the help page"))
    self.helpPushButton.setText(_translate("DataHierarchyEditorDialogBase", "Help"))
    self.cancelPushButton.setToolTip(_translate("DataHierarchyEditorDialogBase", "Close the editor"))
    self.cancelPushButton.setText(_translate("DataHierarchyEditorDialogBase", "Cancel"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    DataHierarchyEditorDialogBase = QtWidgets.QWidget()
    ui = Ui_DataHierarchyEditorDialogBase()
    ui.setupUi(DataHierarchyEditorDialogBase)
    DataHierarchyEditorDialogBase.show()
    sys.exit(app.exec())



"""Input and output functions towards the .eln file-format"""
import os, json, shutil, logging, hashlib, copy
from typing import Any, Optional
from pathlib import Path
from datetime import datetime
from zipfile import ZipFile, ZIP_DEFLATED
import requests
from anytree import Node
from pasta_eln import __version__
from .backend import Backend
from .miscTools import createDirName, generic_hash

# to discuss
# - genre:docType, simulation, experiment/measurement;  status = Done, finished
# - cathegory: project
# - root entry: authors list, single: @id; multiple authors
#    - add several authors
#    - one creator, multiple authors
# - where to store additional metadata, not in ro-crate-metadata, separate files for each entry?
#    - https://github.com/TheELNConsortium/TheELNFileFormat/issues/58
# - how to store different versions?
#    - history: last version
#    - is based based on, ro-crate id OR update action
# - how should the folder structure be? kadi4mat, sampleDB, does-not-matter:
# - sampleDB ro-crate.json: @type:comment!
#   - ??
# - in "ro-crate-metadata.json" / "sdPublisher": "@id": or name
# - how to verify the import
#   - import - export = the same

# Always use RO-crate names
# GENERAL TERMS IN ro-crate-metadata.json (None implies definitely should not be saved)
pasta2json:dict[str,Any] = {
  '_id'         : 'identifier',
  '_rev'        : None,
  '_attachments': None,
  '-attachment' : None,
  '-branch'     : None,
  '-client'     : None,
  '-date'       : 'dateModified',
  '-name'       : 'name',
  '-tags'       : 'keywords',
  'image'       : None,
  'comment'     : 'description',
  'content'     : 'text',
  'links'       : 'mentions',
  'shasum'      : None,
}
json2pasta:dict[str,Any] = {v:k for k,v in pasta2json.items() if v is not None}


# Special terms in other ELNs: only add the ones that are required for function for PASTA
specialTerms:dict[str,dict[str,str]] = {
    'elabFTW': {
      # 'elabid':'_id',
      # 'title':'-name',
      # 'tags':'-tags',
      # 'lastchange':'-date',
      # 'userid':'-user',
      # 'metadata':'metaUser',
      # 'id':'_id',
      # 'category': '-type',
      # 'dateCreated': '-date'
    },
    'SampleDB':{
      'comment':'userComment',
      'genre': '-type'
    }
}


renameELN = {
  'https://kadi.iam.kit.edu':'Kadi4Mat'
}


def tree(graph:dict[Any,Any]) -> str:
  """
  use metadata to create hierarchical tree structure in ascii

  Args:
    graph (dict): tree-graph to be plotted

  Returns:
    str: output of graph as a nice tree
  """
  METADATA_FILE = 'ro-crate-metadata.json'

  def process_part(part:dict[Any,Any], level:int) -> str:
    """
    recursive function call to process this node of the tree-graph

    Args:
      part (dict): dictionary entry of this hasPart=part of a parent node; is a node by-itself
      level (int): hierarchy level

    Returns:
      str: one line in the tree
    """
    prefix = '    '*level + '- '
    # find next node to process
    newNode:list[Any] = [i for i in graph if '@id' in i and i['@id'] == part['@id']]
    if len(newNode) == 1:
      output = f'{prefix} {newNode[0]["@id"]} type:{newNode[0]["@type"]} items: {len(newNode[0])-1}\n'  # -1 because @id is not counted
      subparts = newNode[0].pop('hasPart') if 'hasPart' in newNode[0] else []
      if len(subparts) > 0:  # don't do if no subparts: measurements, ...
        for subpart in subparts:
          output += process_part(subpart, level + 1)
    else:                                                       #not found
      output = f'  items: {len(part) - 1}\n'      # -1 because @id is not counted
    return output

  # main tree-function
  #   find information from master node
  ro_crate_node = [i for i in graph if i["@id"] == METADATA_FILE][0]
  output = '- '+METADATA_FILE+'\n'
  if 'sdPublisher' in ro_crate_node:
    name = ro_crate_node['sdPublisher'].get('name','---')
    output += '    - publisher: ' + name  + '\n'
  if 'version' in ro_crate_node:
    output += '    - version: ' + ro_crate_node['version'] + '\n'
  main_node = [i for i in graph if i["@id"] == "./"][0]
  output += '- ./\n'
  #   iteratively go through list
  for part in main_node['hasPart']:
    output += process_part(part, 1)
  return output


def importELN(backend:Backend, elnFileName:str) -> str:
  '''
  import .eln file from other ELN or from PASTA

  Args:
    backend (backend): backend
    elnFileName (str): name of file

  Returns:
    str: success message
  '''
  elnName = ''
  elnVersion = ''
  projID, projPWD= '', Path('')
  with ZipFile(elnFileName, 'r', compression=ZIP_DEFLATED) as elnFile:
    files = elnFile.namelist()
    logging.info('All files '+', '.join(files))
    dirName=Path(files[0]).parts[0]
    if f'{dirName}/ro-crate-metadata.json' not in files:
      print('**ERROR: ro-crate does not exist in folder. EXIT')
      return '**ERROR: ro-crate does not exist in folder. EXIT'
    graph = json.loads(elnFile.read(f'{dirName}/ro-crate-metadata.json'))["@graph"]
    #find information from master node
    rocrateNode = [i for i in graph if i["@id"].endswith("ro-crate-metadata.json")][0]
    if 'sdPublisher' in rocrateNode:
      elnName     = rocrateNode['sdPublisher'].get('name', '')
      if elnName == '':
        elnName     = rocrateNode['sdPublisher'].get('@id', '')
    elnVersion = rocrateNode['version'] if 'version' in rocrateNode else ''
    logging.info('Import %s %s', elnName, elnVersion)
    if elnName!='PASTA ELN':
      if elnName in specialTerms:
        json2pasta.update(specialTerms[elnName])
    logging.info('ELN and translator: %s %s', elnName, str(json2pasta))
    mainNode    = [i for i in graph if i["@id"]=="./"][0]

    ################
    # subfunctions #
    ################
    def json2pastaFunction(inputData:dict[str,Any]) -> tuple[dict[str,Any], str, list[Any], str]:
      """
      convert json data to pastaStyle

      Args:
        inputData (dict): input data in document

      Returns:
        dict: output data as document
        str: id in eln=path in zip file
        list: list of children
        str: data or dataset
      """
      output = {}
      elnID = inputData['@id'][2:] if inputData['@id'][:2]=='./' else inputData['@id']
      children = inputData.pop('hasPart') if 'hasPart' in inputData else []
      dataType = inputData['@type']
      for key, value in inputData.items():
        if key in ['@id','@type','hasPart','author','contentSize', 'sha256']:
          continue
        if key in json2pasta:  #use known translation
          output[json2pasta[key]] = value
        else:                  #keep name, only prevent causes for errors
          key = f'imported_{key}' if key.startswith(('-','_')) else key
          output[key] = value
      return output, elnID, children, dataType


    def processPart(part:dict[str,str]) -> int:
      """
      recursive function call to process this node

      Args:
        part (dict): dictionary containing this data content

      Returns:
        bool: success of this function
      """
      addedDocs = 1
      if not isinstance(part, dict): #leave these tests in since other .elns might do funky stuff
        print("**ERROR in part",part)
        return False
      logging.info('\nProcess: '+part['@id'])
      # find next node to process
      docS = [i for i in graph if '@id' in i and i['@id']==part['@id']]
      if len(docS)!=1 or backend.cwd is None:
        print('**ERROR zero or multiple nodes with same id', docS,' or cwd is None in '+part['@id'])
        return -1
      doc, elnID, children, dataType = json2pastaFunction(copy.deepcopy(docS[0]))
      if elnName == 'PASTA ELN' and elnID.startswith('http') and ':/' in elnID:
        fullPath = None
      elif elnName == 'PASTA ELN':
        fullPath = backend.basePath/elnID
      else:
        fullPath = backend.basePath/backend.cwd/elnID.split('/')[-1]
      datasetIsFolder = False
      if dataType.lower()=='dataset':
        #print('  Dataset',elnID,dataType)
        if elnName == 'PASTA ELN':
          supplementalInfo = Path(dirName)/elnID/'metadata.json'
        elif elnName == 'eLabFTW':
          supplementalInfo = Path(dirName)/elnID/'export-elabftw.json'
        else:
          logging.info('No additional information in %s', elnName)
          supplementalInfo = Path('')
        if supplementalInfo.as_posix() in elnFile.namelist():
          datasetIsFolder = True
          with elnFile.open(supplementalInfo.as_posix()) as fIn:
            jsonContent = json.loads( fIn.read() )
            if isinstance(jsonContent, list):
              jsonContent = jsonContent[0]
            if elnName == 'PASTA ELN':
              doc.update( jsonContent['__main__'] )
            else:
              doc[f'from {elnName}'] = jsonContent
      if not datasetIsFolder:
        #print('  ELSE ',elnID,dataType)
        if elnName == 'PASTA ELN':
          if elnID.endswith('data_hierarchy.json'):
            return 0
          parentIDList = [i['@id'] for i in graph if 'hasPart' in i and {'@id':'./'+elnID} in i['hasPart']]
          if not parentIDList: #if no parent found, aka for remote data
            parentIDList = [i['@id'] for i in graph if 'hasPart' in i and {'@id':elnID} in i['hasPart']]
          if len(parentIDList)==1:
            metadataPath = Path(dirName)/parentIDList[0]/'metadata.json'
            with elnFile.open(metadataPath.as_posix()) as fIn:
              metadataContent = json.loads( fIn.read() )
              doc.update( metadataContent[doc['_id']] )
        elif elnName == 'eLabFTW' and elnID.endswith('export-elabftw.json'):
          return 0
        elif fullPath is not None:
          if f'{dirName}/{elnID}' in elnFile.namelist():  #could be directory, nothing to copy then
            target = open(fullPath, "wb")
            source = elnFile.open(f'{dirName}/{elnID}')
            with source, target:  #extract one file to its target directly
              shutil.copyfileobj(source, target)
      # save
      # ALL ELNS
      if '-tags' not in doc:
        doc['-tags'] = []
      else:
        doc['-tags'] = [i.strip() for i in doc['-tags'].split(',')]
      # PASTA_ELN
      if elnName == 'PASTA ELN':
        doc['-user'] = '_'
        if datasetIsFolder and fullPath is not None:
          fullPath.mkdir(exist_ok=True)
          with open(fullPath/'.id_pastaELN.json', 'w', encoding='utf-8') as fOut:
            fOut.write(json.dumps(doc))
        elif fullPath is not None:
          if not fullPath.parent.is_dir():
            fullPath.parent.mkdir()
          if f'{dirName}/' + part['@id'][2:] in files:  #if a file is saved
            target = open(fullPath, "wb")
            source = elnFile.open(f'{dirName}/' + part['@id'][2:])
            with source, target:  #extract one file to its target directly
              shutil.copyfileobj(source, target)
            backend.useExtractors(fullPath, doc['shasum'], doc)
          else:
            logging.warning('  could not read file from zip: %s',part['@id'])
        else:
          backend.useExtractors(Path(elnID), doc['shasum'], doc)
        backend.db.saveDoc(doc)
      else:  # OTHER VENDORS
        if dataType.lower()=='dataset':
          docType = 'x'+str(len(elnID.split('/')) - 1)
          nonlocal projID
          nonlocal projPWD
          if docType == 'x1' and projID == '':
            print(f'  Want to created project name:{renameELN.get(elnName, elnName)}')
            projID = backend.addData('x0', {'-name':f'Imported project of {renameELN.get(elnName, elnName)} '})
            backend.changeHierarchy(projID)
            projPWD= Path(backend.cwd)
            backend.hierStack = [projID]
          elif docType == 'x1':
            backend.cwd = Path(projPWD)
            backend.hierStack = [projID]
          # backend.cwd = backend.basePath / Path(elnID).parent
          # if not backend.cwd.exists():
          #   backend.cwd = backend.cwd.parent
        else:
          docType = '-'
        if docType=='x0':
          backend.hierStack = []
        if '_id' in doc:
          doc['externalID'] = doc.pop('_id')
        print(f'Want to add doc:{doc} with type:{docType} and cwd:{backend.cwd}')
        docID = backend.addData(docType, doc)
        if docID[0]=='x':
          backend.hierStack += [docID]
          backend.cwd        = backend.basePath/ backend.db.getDoc(docID)['-branch'][0]['path']

      # children, aka recursive part
      logging.info('subparts:'+', '.join(['  '+i['@id'] for i in children]))
      for child in children:
        if child['@id'].endswith('/metadata.json') or child['@id'].endswith('_metadata.json'):  #skip own metadata
          continue
        addedDocs += processPart(child)

      return addedDocs

    ######################
    #main function
    #iteratively go through list
    addedDocuments = 0
    for part in mainNode['hasPart']:
      if not part['@id'].endswith('ro-crate-metadata.json'):
        addedDocuments += processPart(part)
  #return to home stack and path
  backend.cwd = Path(backend.basePath)
  backend.hierStack = []
  print(f'\n\nGraph in metadata file\n{tree(graph)}')
  return f'Success: imported {str(addedDocuments)} documents from file {elnFileName} from ELN {elnName} {elnVersion}'




##########################################
###               EXPORT               ###
##########################################
def exportELN(backend:Backend, projectID:str, fileName:str='', dTypes:list[str]=[], verbose:bool=True) -> str:
  """
  export eln to file

  Args:
    backend (backend): PASTA backend instance
    projectID (str): docId of project
    fileName (str): fileName which to use for saving; default='' saves in local folder
    dTypes (list): list of strings which should be included in the output, alongside folders x0 & x1; empty list=everything is exported
    verbose (bool): verbose

  Returns:
    str: report of exportation
  """
  # define initial information
  docProject = backend.db.getDoc(projectID)
  dirNameProject = docProject['-branch'][0]['path']
  fileName = fileName or dirNameProject
  fileName = fileName if fileName.endswith('.eln') else f'{fileName}.eln'
  keysInSupplemental:set[str] = set()
  filesNotInProject = []

  def separate(doc: dict[str,Any]) -> tuple[str, dict[str,Any], dict[str,Any]]:
    """ separate document into
    - main information (for all elns) and
    - supplemental information (specific to PastaELN)
    """
    path =  f"{doc['-branch'][0]['path']}/" if doc['-type'][0][0]=='x' else doc['-branch'][0]['path']
    pathUsed = True
    if path is None:
      pathUsed = False
      path = './'+dirNameProject+'/'+doc['_id']
    if not path.startswith(dirNameProject) and not path.startswith('http') and pathUsed:
      filesNotInProject.append(path)
    if not path.startswith('http') and pathUsed:
      path = f'./{path}'
    docMain= {'@id': path}
    docSupp = {}
    for key, value in doc.items():
      if key in pasta2json and pasta2json[key] is not None:
        docMain[pasta2json[key]] = value
      else:
        docSupp[key] = value
    # clean individual entries of docSupp: remove personal data
    del docSupp['-user']
    if 'image' in docSupp:
      del docSupp['image']
    if '_attachments' in docSupp:
      del docSupp['_attachments']
    docSupp['_id'] = docMain['identifier'] #ensure id is present in main and supplementary information
    # clean individual entries of docMain
    if 'keywords' in docMain:
      docMain['keywords'] = ','.join(docMain['keywords'])
    docMain = {k:v for k,v in docMain.items() if v}
    nonlocal keysInSupplemental
    keysInSupplemental = keysInSupplemental.union(docSupp)
    return path, docMain, docSupp


  def appendDocToGraph(graph: list[dict[str,Any]], doc: dict[str,Any]) -> None:
    """
    Append a document / dictionary to a graph

    Args:
      graph (list): graph to be appended to
      doc   (dict): document to append to graph
    """
    idsInGraph = [i['@id'] for i in graph]
    if doc['@id'] not in idsInGraph:
      graph.append(doc)
    return


  def iterateTree(nodeHier:Node, graph:list[dict[str,Any]]) -> Optional[str]:
    """
    Recursive function to translate the hierarchical node into a tree-node

    Args:
      nodeHier (Anytree.Node): anytree node
      graph    (list): list of nodes

    Returns:
      str: tree node
    """
    # separate into main and supplemental information
    doc = backend.db.getDoc(nodeHier.id)
    path, docMain, docSupp = separate(doc)

    hasPart = []
    if (nodeHier.docType[0] not in dTypes) and nodeHier.docType[0][0]!='x' and len(dTypes)>0:
      return None
    for child in nodeHier.children:
      res = iterateTree(child, graph)
      if res is not None:
        hasPart.append( res )
    if nodeHier.id[0]=='x':
      pathMetadata = f'{path}metadata.json'
      docMain['hasPart'] = [{'@id':pathMetadata}]
      if hasPart:
        docMain['hasPart'] += [{'@id':i} for i in hasPart]

      metadata = {'__main__':docSupp}
      hierStack = docSupp['-branch'][0]['stack']+[docSupp['_id']]
      viewFull = backend.db.getView('viewHierarchy/viewHierarchyAll', startKey=' '.join(hierStack))
      viewIDs = [i['id'] for i in viewFull if len(i['key'].split())==len(hierStack)+1 and i['value'][1][0][0]!='x']
      for childID in viewIDs:
        _, _, docChildSupp = separate(backend.db.getDoc(childID))
        metadata[childID] = docChildSupp
      zipContent = json.dumps(metadata)
      roCrateMetadata = {'@id':pathMetadata,
                         '@type':'File',
                         'name':'export_metadata.json',
                         'description':'JSON export',
                         'encodingFormat':'application/json',
                         'contentSize':str(len(zipContent)),
                         'sha256':hashlib.sha256(zipContent.encode()).hexdigest(),
                         'dateModified': docMain['dateModified']
                         }
      elnFile.writestr(f'{dirNameProject}/{pathMetadata[2:]}', zipContent)
      docMain['@type'] = 'Dataset'
      appendDocToGraph(graph, roCrateMetadata)

    fullPath = backend.basePath/path
    if path is not None and fullPath.exists() and fullPath.is_file():
      with open(fullPath, 'rb') as fIn:
        fileContent = fIn.read()
        docMain['contentSize'] = str(len(fileContent))
        docMain['sha256']      = hashlib.sha256(fileContent).hexdigest()
      docMain['@type'] = 'File'
    elif path.startswith('http'):
      response = requests.get(path.replace(':/','://'))
      if response.ok:
        docMain['contentSize'] = str(response.headers['content-length'])
        docMain['sha256']      = hashlib.sha256(response.content).hexdigest()
      else:
        print(f'Info: could not get file {path}')
      docMain['@type'] = 'File'
    elif '@type' not in docMain:  #samples will be here
      docMain['@type'] = 'Dataset'
      docMain['@id'] = docMain['@id'] if docMain['@id'].endswith('/') else f"{docMain['@id']}/"

    appendDocToGraph(graph, docMain)
    return docMain['@id']

  # == MAIN FUNCTION ==
  logging.info('Create eln file %s',fileName)
  with ZipFile(fileName, 'w', compression=ZIP_DEFLATED) as elnFile:
    # ------- Create main graph -------------------
    listHier = backend.db.getHierarchy(projectID, allItems=False)
    graph: list[dict[str,Any]] = []
    iterateTree(listHier, graph)  # create json object from anytree

    # ------------------- create ro-crate-metadata.json header -----------------------
    index:dict[str,Any] = {}
    index['@context']= 'https://w3id.org/ro/crate/1.1/context'
    # master node ro-crate-metadata.json
    graphMaster:list[dict[str,Any]] = []
    graphMisc:list[dict[str,Any]] = []
    masterNodeInfo = {
        '@id': 'ro-crate-metadata.json',
        '@type': 'CreativeWork',
        'about': {
            '@id': './'
        },
        'conformsTo': {
            '@id': 'https://w3id.org/ro/crate/1.1'
        },
        'schemaVersion': 'v1.0',
        'dateCreated': datetime.now().isoformat(),
        'sdPublisher': {
            '@type': 'Organization',
            'name': 'PASTA ELN',
            'logo':
            'https://raw.githubusercontent.com/PASTA-ELN/desktop/main/pasta.png',
            'slogan': 'The favorite ELN for experimental scientists',
            'url': 'https://github.com/PASTA-ELN/',
            'description': f'Version {__version__}',
        },
        'version': '1.0',
        'datePublished':datetime.now().isoformat(),
    }
    graphMaster.append(masterNodeInfo)
    authors = backend.configuration['authors']
    masterNodeRoot = {
        '@id': './',
        '@type': 'Dataset',
        'hasPart': [
            {'@id': f'./{dirNameProject}/'}
        ],
        'name': 'Exported from PASTA ELN',
        'description': 'Exported content from PASTA ELN',
        'license':'CC BY 4.0',
        'author': [{
            'firstName': authors[0]['first'],
            'surname': authors[0]['last'],
            'title': authors[0]['title'],
            'emailAddress': authors[0]['email'],
            'ORCID': authors[0]['orcid'],
            'affiliation': [{
                'organization': authors[0]['organizations'][0]['organization'],
                'RORID': authors[0]['organizations'][0]['rorid'],
            }],
        }],
        'datePublished':datetime.now().isoformat(),
    }
    graphMaster.append(masterNodeRoot)
    zipContent = json.dumps(backend.db.getDoc('-dataHierarchy-'))
    dataHierarchyInfo = {
        '@id':  f'./{dirNameProject}/data_hierarchy.json',
        '@type': 'File',
        'name':  'data structure',
        'description': 'data structure / schema of the stored data',
        'contentSize':str(len(zipContent)),
        'sha256':hashlib.sha256(zipContent.encode()).hexdigest(),
        'datePublished': datetime.now().isoformat()
    }
    graphMisc.append(dataHierarchyInfo)
    graph[-1]['hasPart'] += [{'@id': f'./{dirNameProject}/data_hierarchy.json'}]
    elnFile.writestr(f'{dirNameProject}/{dirNameProject}/data_hierarchy.json', zipContent)

    # ------------------ copy data-files --------------------------
    # datafiles are already in the graph-graph: only copy and no addition to graph
    for path, _, files in os.walk(backend.basePath/dirNameProject):
      if '/.git' in path:  #if use datalad
        continue
      relPath = os.path.relpath(path, backend.basePath) #path of the folder
      for iFile in files:                         #iterate through all files in folder
        if iFile.startswith('.git') or iFile=='.id_pastaELN.json':
          continue
        elnFile.write(f'{path}/{iFile}', f'{dirNameProject}/{relPath}/{iFile}')
    for path in set(filesNotInProject):  #set ensures that only added once
      elnFile.write(str(backend.basePath/path), f'{dirNameProject}/{path}')

    #finalize file
    index['@graph'] = graphMaster+graph+graphMisc
    elnFile.writestr(f'{dirNameProject}/ro-crate-metadata.json', json.dumps(index))
    if verbose:
      print(json.dumps(index, indent=3))
  # end writing zip file
  # temporary json output
  # with open(fileName[:-3]+'json','w', encoding='utf-8') as fOut:
  #   fOut.write( json.dumps(index, indent=2) )
  keysInSupplemental = {i for i in keysInSupplemental if i not in pasta2json}
  logging.info('Keys in supplemental information'+', '.join(keysInSupplemental))
  return f'Success: exported {len(graph)} graph-nodes into file {fileName}'
